import unittest
from cratermaker import Target, Scaling

class TestScale(unittest.TestCase):
   
    def test_scale_from_catalogue(self):
        # Test creating a scaling from the catalogue
        scaling = Scaling.maker(material="Soft Rock")
        self.assertEqual(scaling.K1, 0.20)
        self.assertEqual(scaling.mu, 0.55)
        self.assertEqual(scaling.Ybar, 7.60e6)
        self.assertEqual(scaling.target.density, 2250.0)
        self.assertEqual(scaling.projectile.density, 2250.0)
        self.assertIsInstance(scaling.K1, float)
        self.assertIsInstance(scaling.mu, float)
        self.assertIsInstance(scaling.Ybar, float)
        self.assertIsInstance(scaling.target.density, float)
        self.assertIsInstance(scaling.projectile.density, float)
        return

    def test_scale_override_catalogue(self):
        # Test overriding a property from the catalogue
        scaling = Scaling.maker(material="Soft Rock", Ybar=0.0)
        self.assertEqual(scaling.Ybar, 0.0)
        return

    def test_custom_scale(self):
        # Test creating a custom scaling
        scaling = Scaling.maker(material="Flubber", K1=3.8, mu=0.1, Ybar=1e7, density=2000.0)
        self.assertEqual(scaling.K1, 3.8)
        self.assertEqual(scaling.mu, 0.1)
        self.assertEqual(scaling.Ybar, 1e7)
        self.assertEqual(scaling.target.density, 2000.0)
        return

    def test_incomplete_custom_scale(self):
        # Test incomplete custom scaling creation
        with self.assertRaises(ValueError):
            Scaling.maker(material="Flubber", density=2000.0)
        with self.assertRaises(ValueError):
            Scaling.maker(material="Blorp", mu=0.1, Ybar=1e7, density=2000.0)
        return
    
    def test_scale_override_catalogue(self):
        # Test overriding a property from the catalogue
        target = Target(name="Mars")
        scaling = Scaling.maker(target=target,material="Sand")
        self.assertEqual(scaling.material, "Sand")    

if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, mock_open
from cratermaker.utils.general_utils import _set_properties, parameter, _create_catalogue

mock_properties = [
    "name",       "prop",     "param"
]
mock_values = [
    ("foo",     1.00,     2.00),
    ("bar",     3.00,     4.00),
]  
class Dummy:
    def __init__(self):
        self._name = None
        self._prop = None
        self._param = None
        self._user_defined = set()

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, value):
        self._name = value

    @property
    def prop(self):
        return self._prop

    @prop.setter
    def prop(self, value):
        self._prop = value

    @parameter
    def param(self):
        return self._param

    @param.setter
    def param(self, value):
        self._param = value

    @property
    def catalogue_key(self):
        return "name"

class TestGeneralUtils(unittest.TestCase):
    def setUp(self):
        # Set up any initial configurations or objects you need before each test
        self.dummy = Dummy()
        return
        
    def test_kwargs(self):
        _set_properties(self.dummy, name="baz", prop=5.0, param=6.0)
        self.assertEqual(self.dummy.name, "baz")
        self.assertEqual(self.dummy.prop, 5.0)
        self.assertEqual(self.dummy.param, 6.0)
       
        return    
    
    def test_catalogue(self):
        # Test a regular catalogue with multiple entries (must be selected by  name)
        catalogue = _create_catalogue(mock_properties, mock_values)
        for k,v in catalogue.items():
            _set_properties(self.dummy, catalogue=catalogue, name=k)
            self.assertEqual(self.dummy.name, k)
            self.assertEqual(self.dummy.prop, v['prop'])
            self.assertEqual(self.dummy.param, v['param'])
            
        # Test a single entry catalogue (name can be omitted)
        name = next(iter(catalogue))
        _set_properties(self.dummy, catalogue={name:catalogue[name]})
    
        # Create an invalid catalogue structure
        invalid_catalogue = {"invalid_key": "invalid_value"}
        
        # Test if ValueError is raised for invalid catalogue structure or other problems
        with self.assertRaises(ValueError):  
            _set_properties(self.dummy, catalogue="bad_value") # Not even a dict
        with self.assertRaises(ValueError):  
            _set_properties(self.dummy, catalogue=invalid_catalogue, key="invalid_key") # Not a nested dict
        
        return


    def test_kwarg_override(self):
        param_new = 42.0
        catalogue = _create_catalogue(mock_properties, mock_values)
        name = next(iter(catalogue))
        _set_properties(self.dummy, catalogue=catalogue, key=name, param=param_new)
        self.assertEqual(self.dummy.name, name)
        self.assertEqual(self.dummy.prop, catalogue[name]['prop'])
        self.assertEqual(self.dummy.param, param_new)
        
        # test that overrides are ignored when None is passed:
        _set_properties(self.dummy, catalogue=catalogue, key=name, param=None) 
        self.assertEqual(self.dummy.param, catalogue[name]['param'])
        return

if __name__ == '__main__':
    unittest.main()



"""
Plot a distribution of random impact angles
===========================================

.. rubric:: By David Minton

This example demonstrates the generation of random impact angles.

"""

import matplotlib.pyplot as plt
import numpy as np

from cratermaker.utils.montecarlo_utils import get_random_impact_angle

# Sample data generation
size = 10000
angles = get_random_impact_angle(size=size)

# Number of bins
bins = 50
observed_counts, bins_ang = np.histogram(angles, bins=bins, range=(0.0, 90.0))

# Calculate expected distribution
uniform_dist = np.linspace(0, 1, size)
transformed_angles = np.rad2deg(np.arcsin(np.sqrt(uniform_dist)))
expected_counts, _ = np.histogram(transformed_angles, bins=bins, range=(0.0, 90.0))

# Plotting
fig, ax = plt.subplots(figsize=(8, 4))

# Observed counts
ax.bar(
    bins_ang[:-1],
    observed_counts,
    width=np.diff(bins_ang),
    align="edge",
    label="Observed",
    alpha=0.5,
)

# Expected counts
ax.plot(bins_ang[:-1], expected_counts, label="Expected", color="red")

ax.set_xlabel("Impact Angle (deg)")
ax.set_ylabel("Count")
ax.legend()
ax.set_title("Impact Angle Distribution")

plt.tight_layout()
plt.show()



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any

import numpy as np
from numpy.random import Generator

from cratermaker.constants import FloatLike
from cratermaker.utils.component_utils import ComponentBase, import_components

if TYPE_CHECKING:
    from cratermaker.components.projectile import Projectile
    from cratermaker.components.target import Target


class Scaling(ComponentBase):
    _registry: dict[str, Scaling] = {}

    """
    This is the abstract base class for all scaling models. It defines the interface for converting between projectile and crater diameters.

    Parameters
    ----------
    target : Target | str, default="Moon"
        The target body for the impact. Can be a Target object or a string representing the target name.
    projectile : Projectile | str, default="asteroids"
        The projectile model for the impact. Can be an Projectile object or a string representing the projectile name.
    rng : numpy.random.Generator | None
        A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        The rng_rng_seed for the RNG. If None, a new RNG is created.
    rng_state : dict, optional
        The state of the random number generator. If None, a new state is created.
    **kwargs : Any
        Additional keyword arguments.
    """

    def __init__(
        self,
        target: Target | str | None = None,
        projectile: Projectile | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        from cratermaker.components.projectile import Projectile
        from cratermaker.components.target import Target

        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        object.__setattr__(self, "_target", None)
        object.__setattr__(self, "_projectile", None)
        # combine the kwargs with the common_args, giving common_args priority
        kwargs = {**kwargs, **vars(self.common_args)}
        self._target = Target.maker(target, **kwargs)
        self._projectile = Projectile.maker(projectile, target=self.target, **kwargs)

    @abstractmethod
    def projectile_to_transient(self, **kwargs: Any) -> np.float64: ...
    @abstractmethod
    def transient_to_projectile(self, **kwargs: Any) -> np.float64: ...
    @abstractmethod
    def transient_to_final(
        self, transient_diameter: FloatLike
    ) -> tuple[np.float64, str]: ...
    @abstractmethod
    def final_to_transient(
        self, final_diameter: FloatLike, morphology_type: str | None = None, **kwargs
    ) -> np.float64: ...
    @abstractmethod
    def recompute(self, **kwargs: Any) -> None: ...

    @classmethod
    def maker(
        cls,
        scaling: str | Scaling | None = None,
        target: Target | str | None = None,
        projectile: Projectile | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ) -> Scaling:
        """
        Initialize a scaling model based on the provided name or class.

        Parameters
        ----------
        scaling : str, Scaling, or None, default=None
            The name of the scaling model to initialize. If None, the default model is used.
        target : Target | str, default="Moon"
            The target body for the impact. Can be a Target object or a string representing the target name.
        projectile : Projectile | str, default="asteroids"
            The projectile model for the impact. Can be an Projectile object or a string representing the projectile name.
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Scaling
            An instance of the specified scaling model.

        Raises
        ------
        KeyError
            If the specified scaling model name is not found in the registry.
        TypeError
            If the specified scaling model is not a string or a subclass of Scaling.
        """

        if scaling is None:
            scaling = "montecarlo"
        scaling = super().maker(
            component=scaling,
            target=target,
            projectile=projectile,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )
        if projectile is not None:
            scaling.projectile = projectile
        if target is not None:
            scaling.target = target

        return scaling

    def __str__(self) -> str:
        base = super().__str__()
        return (
            f"{base}\n"
            f"Target: {self.target._component_name}\n"
            f"Projectile: {self.projectile._component_name}"
        )

    @property
    def target(self):
        """
        The target body for the impact.

        Returns
        -------
        Target
        """
        return self._target

    @target.setter
    def target(self, value):
        from cratermaker.components.target import Target

        self._target = Target.maker(value, **vars(self.common_args))
        return

    @property
    def projectile(self):
        """
        The projectile model for the impact.

        Returns
        -------
        Projectile
        """
        return self._projectile

    @projectile.setter
    def projectile(self, value):
        from cratermaker.components.projectile import Projectile

        self._projectile = Projectile.maker(value, **vars(self.common_args))
        return

    @property
    def model(self):
        """
        The name of the scaling model.

        Returns
        -------
        str
        """
        return self._component_name


import_components(__name__, __path__, ignore_private=True)



"""
Plot a distribution of random sizes
===================================

.. rubric:: By David Minton

This example demonstrates the generation of random impact sizes, which is used to sample production function SFDs. In this example, we just draw from a simple power law.

"""

import matplotlib.pyplot as plt
import numpy as np

from cratermaker.utils.montecarlo_utils import get_random_size

num_realizations = 100
nbins = 10
size = 100000
Dhi = 100.0
p = 2.0
C = Dhi**p
Dlo = (size / C) ** (-1.0 / p)
diameters = np.exp(np.linspace(np.log(Dlo), np.log(100 * Dhi), nbins))
# Test a simple power law SFD
cdf = C * diameters**-p

fig = plt.figure(figsize=(4, 3))
ax = fig.add_subplot(1, 1, 1)

ax.set_xlabel("$D$")
ax.set_ylabel("$N_{>D}$")
ax.set_xscale("log")
ax.set_yscale("log")
ax.set_ylim([1.0, 1.1 * size])

for i in range(num_realizations):
    new_diameters = get_random_size(diameters, cdf, mu=size)
    new_diameters = np.sort(new_diameters)[::-1]
    nval = np.arange(1, new_diameters.size + 1)
    if i == 0:
        label = "Sampled SFD"
    else:
        label = None

    ax.plot(new_diameters, nval, label=label, alpha=0.25, color="cornflowerblue")

ax.plot(diameters, cdf, label="Model SFD", color="black")
ax.legend(loc="upper right")

plt.tight_layout()
plt.show()



from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np
from numpy.random import BitGenerator, Generator, RandomState, SeedSequence
from numpy.typing import ArrayLike

from cratermaker.utils.general_utils import parameter


@dataclass
class CommonArgs:
    simdir: Path
    rng: Generator | None
    rng_seed: int | None
    rng_state: dict | None


class CratermakerBase:
    def __init__(
        self,
        simdir: str | Path | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs,
    ):
        """
        Initialize the CratermakerBase class.

        Parameters
        ----------
        simdir : str | Path
            The main project simulation directory. Defaults to the current working directory if None.
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.
        """
        object.__setattr__(self, "_user_defined", set())
        object.__setattr__(self, "_rng", None)
        object.__setattr__(self, "_rng_seed", None)
        object.__setattr__(self, "_rng_state", None)
        object.__setattr__(self, "_simdir", None)
        self.simdir = simdir

        self._rng_seed = rng_seed
        self.rng, self.rng_state = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state)

        super().__init__()

    def to_config(self, remove_common_args: bool = False, **kwargs: Any) -> dict[str, Any]:
        """
        Converts values to types that can be used in yaml.safe_dump. This will convert various types into a format that can be saved in a human-readable YAML file.

        Parameters
        ----------
        obj : Any
            The object whose attributes will be stored.  It must have a _user_defined attribute.
        remove_common_args : bool, optional
            If True, remove the set of common arguments that are shared among all components of the project from the configuration. Defaults to False.
        **kwargs : Any
            Additional keyword arguments for subclasses.

        Returns
        -------
        dict[str, Any]
            A dictionary of the object's attributes that can be serialized to YAML.

        Notes
        -----
        - The function will ignore any attributes that are not serializable to human-readable YAML. Therefore, it will ignore anything that cannot be converted into a str, int, float, or bool.
        - The function will convert Numpy types to their native Python types.
        """
        return _to_config(self, remove_common_args=remove_common_args, **kwargs)

    @parameter
    def simdir(self):
        """
        The main project simulation directory.

        Returns
        -------
        Path
            The initialized simulation directory as a Path object. Will be a relative path if possible, otherwise will be absolute. If it doesn't exist, it will be created.
        """
        return self._simdir

    @simdir.setter
    def simdir(self, value):
        self._simdir = _simdir_init(value)

    @parameter
    def rng_seed(self):
        """
        The random rng_seed for the simulation RNG.

        Returns
        -------
        int or None
            The integer rng_seed used to initialize the RNG, or None if not set.
        """
        return self._rng_seed

    @rng_seed.setter
    def rng_seed(self, value):
        if value is not None:
            if not isinstance(value, (int | np.integer)) or np.isnan(value) or np.isinf(value) or value < 0:
                raise TypeError("rng_seed must be a positive integer")
            self._rng_seed = int(value)
        else:
            self._rng_seed = None

    @property
    def rng(self):
        """
        The random number generator used for stochastic elements of the simulation.

        Returns
        -------
        numpy.random.Generator or None
            The RNG instance, or None if not initialized.
        """
        return self._rng

    @rng.setter
    def rng(self, value):
        self._rng, _ = _rng_init(rng=value, rng_seed=self.rng_seed, rng_state=self.rng_state)

    @parameter
    def rng_state(self):
        """
        The state of the random number generator.

        Returns
        -------
        dict or None
            A dictionary representing the RNG state, or None if the RNG is not initialized.
        """
        return self.rng.bit_generator.state if self.rng is not None else None

    @rng_state.setter
    def rng_state(self, value):
        _, self._rng_state = _rng_init(rng=self.rng, rng_seed=self.rng_seed, rng_state=value)

    @property
    def common_args(self) -> CommonArgs:
        return CommonArgs(
            simdir=self.simdir,
            rng=self.rng,
            rng_seed=self.rng_seed,
            rng_state=self.rng_state,
        )


def _rng_init(
    rng: Generator | None = None,
    rng_seed: int | ArrayLike | SeedSequence | BitGenerator | Generator | RandomState | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> tuple[Generator, dict]:
    """
    Initialize the random number generator (RNG) based on the provided rng_seed.

    Parameters
    ----------
    rng : Generator, optional
        The random number generator to be initialized.
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        The rng_seed for the RNG. If None, a new RNG is created.
    rng_state : dict, optional
        Set the rng_state of the RNG.

    Returns
    -------
    Generator
        The initialized RNG instance.
    dict
        The state of the RNG.

    Notes
    -----
    - If rng is provided, it will be used directly and any input rng_seed or rng_state will be ignored.
    - If both rng_seed and rng_state are provided, rng_state will take precedence.
    - If neither rng nor rng_seed is provided, a new RNG will be created using the default rng_seed and the state will be returned.
    """
    if rng is not None:
        if not isinstance(rng, Generator):
            raise ValueError("rng must be a numpy.random.Generator instance.")
    else:
        if rng_seed is None:
            rng = np.random.default_rng()
        if rng_state is not None:
            if not isinstance(rng_state, dict):
                raise ValueError("rng_state must be a dictionary.")
            try:
                rng.bit_generator.state = rng_state
            except Exception as e:
                raise ValueError("Invalid rng_state provided.") from e
        elif rng_seed is not None:
            try:
                rng = np.random.default_rng(seed=rng_seed)
            except Exception as e:
                raise ValueError("Invalid rng_seed provided.") from e

    rng_state = rng.bit_generator.state
    return rng, rng_state


def _simdir_init(simdir: str | Path | None = None, **kwargs: Any) -> Path:
    """
    Initialize the simulation directory.

    Parameters
    ----------
    simdir : str | Path | None
        The main project simulation directory. Defaults to the current working directory if None.

    Returns
    -------
    Path
        The initialized simulation directory as a Path object. Will be a relative path if possible, otherwise will be absolute.
    """
    if simdir is None:
        p = Path.cwd()
    else:
        try:
            p = Path(simdir)
            if not p.is_absolute():
                p = Path.cwd() / p
            p.mkdir(parents=True, exist_ok=True)
            p = p.resolve()
        except TypeError as e:
            raise TypeError("simdir must be a path-like object (str, Path, or None)") from e
    try:
        simdir = p.relative_to(Path.cwd())
    except ValueError:
        simdir = p
    return simdir


def _convert_for_yaml(obj):
    if isinstance(obj, dict):
        return {k: _convert_for_yaml(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_convert_for_yaml(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(_convert_for_yaml(v) for v in obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, (np.integer, np.floating)):
        return obj.item()
    elif isinstance(obj, Path):
        return str(obj)
    elif isinstance(obj, (str, int, float, bool)):
        return obj
    elif obj is None:
        return None
    else:
        return str(obj)


def _to_config(obj, remove_common_args: bool = False, **kwargs: Any) -> dict[str, Any]:
    config = _convert_for_yaml({name: getattr(obj, name) for name in obj._user_defined if hasattr(obj, name)})
    if remove_common_args:
        config = {key: value for key, value in config.items() if key not in obj.common_args.__dict__}
    return {key: value for key, value in config.items() if value is not None}



import tempfile
import unittest
from pathlib import Path

import numpy as np
from numpy.random import Generator

from cratermaker.core.base import (
    CommonArgs,
    CratermakerBase,
    _rng_init,
    _simdir_init,
    _to_config,
)


class TestBase(unittest.TestCase):
    def test_rng_argument_validation(self):
        # Invalid rng argument (not a Generator)
        with self.assertRaises(ValueError):
            _rng_init(rng=42)

        # Invalid rng_state type (not a dict)
        with self.assertRaises(ValueError):
            _rng_init(rng_state=123)

        # Invalid rng_seed (object instead of int)
        with self.assertRaises(ValueError):
            _rng_init(rng_seed="baz")

        # Accept random kwargs
        _rng_init(rng=np.random.default_rng(), foo="bar")

    def test_rng_consistency(self):
        # Basic initialization
        rng1, state1 = _rng_init()
        self.assertIsInstance(rng1, Generator)
        self.assertIsInstance(state1, dict)

        _ = rng1.uniform(0, 1)

        # Using the same rng instance
        rng2, _ = _rng_init(rng=rng1)
        _ = rng2.uniform(0, 1)

        # Using a fixed rng_seed
        rng3, _ = _rng_init(rng_seed=42)
        u1 = rng3.uniform(0, 1)
        rng4, _ = _rng_init(rng_seed=42)
        u2 = rng4.uniform(0, 1)
        self.assertEqual(u1, u2)

        # Seed ignored when rng is provided
        rng5, _ = _rng_init(rng=rng3, rng_seed=123)
        u3 = rng5.uniform(0, 1)
        self.assertNotEqual(u3, u1)

        # Restore from state
        rng6, state1 = _rng_init(rng_seed=42)
        u4 = rng6.uniform(0, 1)
        rng7, _ = _rng_init(rng_state=state1)
        u5 = rng7.uniform(0, 1)
        self.assertEqual(u4, u5)

        # Modified state should diverge
        rng8, _ = _rng_init(rng_seed=42)
        _ = rng8.uniform(0, 1)
        rng9, _ = _rng_init(rng=rng8, rng_seed=42)
        u6 = rng9.uniform(0, 1)
        self.assertNotEqual(u4, u6)

    def test_init_simdir(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # Test with None
            newsimdir = _simdir_init(simdir)
            self.assertTrue(newsimdir.is_dir())

            # Test with a string path
            target_path = Path(simdir).resolve()
            newsimdir = _simdir_init(simdir=simdir)
            self.assertTrue(newsimdir.is_dir())
            self.assertEqual(newsimdir, target_path)

            # Test with a Path object
            simdir_path = _simdir_init(simdir=Path(simdir))
            self.assertTrue(simdir_path.is_dir())
            self.assertEqual(simdir_path, target_path)

            # Test with a relative path
            relative_path = "relative_simdir"
            newsimdir = _simdir_init(simdir=relative_path)
            self.assertTrue(newsimdir.is_dir())
            self.assertEqual(str(newsimdir), relative_path)
            newsimdir.rmdir()

            # Test with an invalid path
            with self.assertRaises(TypeError):
                _simdir_init(123)

    def test_to_config(self, **kwargs):
        class Dummy:
            def __init__(self):
                self._user_defined = [
                    "int_attr",
                    "float_attr",
                    "bool_attr",
                    "str_attr",
                    "np_int_attr",
                    "np_float_attr",
                    "np_array_attr",
                    "path_attr",
                    "dict_attr",
                    "list_attr",
                    "tuple_attr",
                    "none_attr",
                    "complex_attr",
                ]
                self.int_attr = 1
                self.float_attr = 2.5
                self.bool_attr = True
                self.str_attr = "hello"
                self.np_int_attr = np.int32(3)
                self.np_float_attr = np.float64(4.5)
                self.np_array_attr = np.array([1, 2, 3])
                self.path_attr = Path("/tmp/test")
                self.dict_attr = {"a": np.int16(10), "b": [np.float32(1.1), "x"]}
                self.list_attr = [1, np.float64(2.2)]
                self.tuple_attr = (3, np.int64(4))
                self.none_attr = None
                self.complex_attr = complex(1, 2)

        dummy = Dummy()
        config = _to_config(dummy)

        expected_keys = {
            "int_attr",
            "float_attr",
            "bool_attr",
            "str_attr",
            "np_int_attr",
            "np_float_attr",
            "np_array_attr",
            "path_attr",
            "dict_attr",
            "list_attr",
            "tuple_attr",
            "complex_attr",
        }

        self.assertEqual(set(config.keys()), expected_keys)
        self.assertEqual(config["int_attr"], 1)
        self.assertEqual(config["float_attr"], 2.5)
        self.assertEqual(config["bool_attr"], True)
        self.assertEqual(config["str_attr"], "hello")
        self.assertEqual(config["np_int_attr"], 3)
        self.assertEqual(config["np_float_attr"], 4.5)
        self.assertEqual(config["np_array_attr"], [1, 2, 3])
        self.assertEqual(config["path_attr"], str(Path("/tmp/test")))
        self.assertEqual(config["dict_attr"]["a"], 10)
        self.assertAlmostEqual(config["dict_attr"]["b"][0], 1.1, places=6)
        self.assertEqual(config["dict_attr"]["b"][1], "x")
        self.assertEqual(config["list_attr"][0], 1)
        self.assertAlmostEqual(config["list_attr"][1], 2.2, places=6)
        self.assertEqual(config["tuple_attr"], (3, 4))
        self.assertEqual(config["complex_attr"], str(complex(1, 2)))
        self.assertNotIn("none_attr", config)

    def test_cratermakerbase_and_commonargs(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # Set up test values
            simdir = Path(simdir).resolve()
            rng, rng_state = _rng_init(rng_seed=123)

            # Test CommonArgs dataclass
            args = CommonArgs(simdir=simdir, rng=rng, rng_seed=123, rng_state=rng_state)
            self.assertEqual(args.simdir, simdir)
            self.assertEqual(args.rng_seed, 123)
            self.assertEqual(args.rng, rng)
            self.assertEqual(args.rng_state, rng_state)

            # Test CratermakerBase initialization
            base = CratermakerBase(
                simdir=simdir, rng=rng, rng_seed=123, rng_state=rng_state
            )

            # Attributes
            self.assertEqual(base.simdir, simdir)
            self.assertEqual(base._rng_seed, 123)
            self.assertEqual(base._rng, rng)  # direct access
            self.assertEqual(base._rng_state, rng_state)
            self.assertIsInstance(base.rng, Generator)
            self.assertEqual(base.rng_state, rng_state)

            # Test RNG reproducibility
            val1 = base.rng.uniform()
            new_rng, _ = _rng_init(rng_state=rng_state)
            val2 = new_rng.uniform()
            self.assertAlmostEqual(val1, val2, places=7)


if __name__ == "__main__":
    unittest.main()



from pathlib import Path
from typing import Any

import numpy as np
from numpy.typing import NDArray

from cratermaker.components.surface import Surface
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike, PairOfFloats
from cratermaker.utils.general_utils import format_large_units, parameter


@Surface.register("arbitrary_resolution")
class ArbitraryResolutionSurface(Surface):
    """
    Create a uniform grid configuration with an arbitrary user-defined pixel size. This will not be as nice as the regular IcosphereSurface, but can be any resolution desired.

    Parameters
    ----------
    pix : float
        The approximate face size for the mesh in meters.
    target : Target, optional
        The target body or name of a known target body for the impact simulation.
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is True.
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False.
    simdir : str | Path
        The main project simulation directory. Defaults to the current working directory if None.

    Returns
    -------
    ArbitraryResolutionSurface
        An instance of the ArbitraryResolutionSurface class initialized with the given pixel size.
    """

    def __init__(
        self,
        pix: FloatLike | None = None,
        target: Target | str | None = None,
        reset: bool = False,
        regrid: bool = False,
        simdir: str | Path | None = None,
        **kwargs: Any,
    ):
        super().__init__(target=target, simdir=simdir, **kwargs)
        self.pix = pix
        self._load_from_files(reset=reset, regrid=regrid, **kwargs)

    def __str__(self) -> str:
        base = super().__str__()
        pix_mean = format_large_units(self.pix_mean, quantity="length")
        pix_std = format_large_units(self.pix_std, quantity="length")
        return f"{base}\nEffective pixel size: {pix_mean} +/- {pix_std}"

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.
        """
        return [self._component_name, self.radius, self.pix]

    @parameter
    def pix(self):
        """
        The approximate face size for a cell of the mesh.
        """
        return self._pix

    @pix.setter
    def pix(self, value: FloatLike):
        if value is None:
            value = (
                np.sqrt(4 * np.pi * self.radius**2) * 1e-3
            )  # Default mesh scale that is somewhat comparable to a 1000x1000 CTEM grid
        elif not isinstance(value, FloatLike) or np.isnan(value) or np.isinf(value) or value <= 0:
            raise TypeError("pix must be a positive float")
        self._pix = float(value)

    @staticmethod
    def _distribute_points(
        distance: FloatLike,
        radius: FloatLike = 1.0,
        lon_range: PairOfFloats = (-180, 180),
        lat_range: PairOfFloats = (-90, 90),
    ) -> NDArray:
        """
        Distributes points on a sphere using Deserno's algorithm (Deserno 2004).

        Parameters
        ----------
        distance : float
            Approximate distance between points, used to determine the number of points, where n = 1/distance**2 when distributed over the whole sphere.
        radius : float, optional
            Radius of the sphere. Default is 1.0
        lon_range : tuple, optional
            Range of longitudes in degrees. Default is (-180,180).
        lat_range : tuple, optional
            Range of latitudes in degrees. Default is (-90,90).

        Returns
        -------
        (3,n) ndarray of np.float64
            Array of cartesian points on the sphere.

        References
        ----------
        - Deserno, Markus., 2004. How to generate equidistributed points on the surface of a sphere. https://www.cmu.edu/biolphys/deserno/pdf/sphere_equi.pdf

        """

        def _sph2cart(theta, phi, r):
            """
            Converts spherical coordinates to Cartesian coordinates.

            Parameters
            ----------
            theta : float
                Inclination angle in radians.
            phi : float
                Azimuthal angle in radians.
            r : float
                Radius.
            """
            x = r * np.sin(theta) * np.cos(phi)
            y = r * np.sin(theta) * np.sin(phi)
            z = r * np.cos(theta)
            return x, y, z

        phi_range = np.deg2rad(lon_range) + np.pi
        theta_range = np.deg2rad(lat_range) + np.pi / 2
        points = []

        n = int(4 * np.pi / distance**2)
        if n < 1:
            return

        a = 4 * np.pi / n
        d = np.sqrt(a)
        mtheta = int(np.round(np.pi / d))
        dtheta = np.pi / mtheta
        dphi = a / dtheta

        thetavals = np.pi * (np.arange(mtheta) + 0.5) / mtheta
        thetavals = thetavals[(thetavals >= theta_range[0]) & (thetavals < theta_range[1])]

        for theta in thetavals:
            mphi = max(1, int(np.round(2 * np.pi * np.sin(theta) / dphi)))
            phivals = 2 * np.pi * np.arange(mphi) / mphi
            phivals = phivals[(phivals >= phi_range[0]) & (phivals < phi_range[1])]
            for phi in phivals:
                points.append(_sph2cart(theta, phi, radius))
        if len(points) == 0:
            return

        points = np.array(points, dtype=np.float64)
        decimals = int(np.ceil(np.log10((2 * np.pi * radius) / distance)))
        points = np.unique(points.T.round(decimals=decimals), axis=0).T
        points = points.T

        return points

    def _generate_face_distribution(self, **kwargs: Any) -> NDArray:
        """
        Creates the points that define the mesh centers.

        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.
        """
        pix_str = format_large_units(self.pix, quantity="length")
        print(f"Generating a mesh with uniformly distributed faces of size ~{pix_str}.")
        points = self._distribute_points(distance=self.pix / self.radius)
        points[:, 0] = np.array([0, 0, 1])
        points[:, -1] = np.array([0, 0, -1])
        return points



from cratermaker.components.projectile import Projectile
from cratermaker.components.scaling import Scaling
from cratermaker.components.scaling.montecarlo import MonteCarloScaling
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike


@Scaling.register("ctem")
class CTEMScaling(MonteCarloScaling):
    """
    This is an operations class for computing the scaling relationships between projectiles and craters.

    This class encapsulates the logic for converting between projectile properties and crater properties,
    as well as determining crater morphology based on size and target propertiesImplements the scaling laws described in Richardson (2009) that were implemented in CTEM.

    Parameters
    ----------
    target : Target | str, default="Moon"
        The target body for the impact. Can be a Target object or a string representing the target name.
    projectile : Projectile | str, default="asteroids"
        The projectile model for the impact. Can be an Projectile object or a string representing the projectile name.
    material : str or None
        Name of the target material composition of the target body to look up from the built-in catalogue. Options include "water", "sand", "dry soil", "wet soil", "soft rock", "hard rock", and "ice".
    K1 : FloatLike, optional
        Variable used in crater scaling (see Richardson 2009)
    mu : FloatLike, optional
        Variable used in crater scaling (see Richardson 2009)
    Ybar : FloatLike, optional
        The strength of the target material, (Pa)
    density : FloatLike, optional
        Volumentric density of target material, (kg/m^3)
    **kwargs : Any
        Additional keyword arguments.

    Notes
    -----
    - The `target` parameter is required and must be an instance of the `Target` class.
    - The `material` parameter is optional. If not provided, it will be retrieved from `target`. Setting it explicitly will override the value in `target`.
    - The `K1`, `mu`, `Ybar`, and `density` parameters are optional. If not provided, they will be retrieved from the material catalogue based on the `material`. Setting them explicitly will override the values in the catalogue.
    - The built-in material property values are from Holsapple (1993) and Kraus et al. (2011).
    - Complex craater scaling parameters are a synthesis of Pike (1980), Croft (1985), and Schenk et al. (2004).

    References
    ----------
    - Richardson, J.E., 2009. Cratering saturation and equilibrium: A new model looks at an old problem. Icarus 204, 697-715. https://doi.org/10.1016/j.icarus.2009.07.029
    - Holsapple, K.A., 1993. The scaling of impact processes in planetary sciences 21, 333-373. https://doi.org/10.1146/annurev.ea.21.050193.002001
    - Kraus, R.G., Senft, L.E., Stewart, S.T., 2011. Impacts onto H2O ice: Scaling laws for melting, vaporization, excavation, and final crater size. Icarus 214, 724-738. https://doi.org/10.1016/j.icarus.2011.05.016
    - Pike, R.J., 1980. Control of crater morphology by gravity and target type - Mars, earth, moon. In: Lunar and Planetary Science Conference 11, 2159-2189.
    - Croft, S.K., 1985. The scaling of complex craters. Proceedings of the Fifteenth Lunar and Planetary Science Conference, Part 2 Journal of Geophysical Research 90, Supplement, C828-C842.
    - Schenk, P.M., Chapman, C.R., Zahnle, K., Moore, J.M., 2004. Ages and interiors: the cratering record of the Galilean satellites, Cambridge University Press. Cambridge University Press, Cambridge, UK.
    """

    def __init__(
        self,
        target: Target | str | None = None,
        projectile: Projectile | str | None = None,
        material: str | None = None,
        K1: FloatLike | None = None,
        mu: FloatLike | None = None,
        Ybar: FloatLike | None = None,
        density: FloatLike | None = None,
        **kwargs,
    ):
        super().__init__(
            target=target,
            projectile=projectile,
            material=material,
            K1=K1,
            mu=mu,
            Ybar=Ybar,
            density=density,
            **kwargs,
        )
        object.__setattr__(self, "_montecarlo_scaling", False)
        self.recompute()  # Recreate this with MC turned off

    def recompute(self):
        """
        Computes and sets the internal attributes for transition factors between simple and complex craters.
        """
        # Constants from CTEM
        CXEXPS = (
            1 / 0.885 - 1.0
        )  # Complex crater scaling explonent for silicate rock (Croft 1985)
        SIMCOMKS = 16533.8  # Simple-to-complex transition scaling coefficient for silicate rock
        SIMCOMPS = (
            -1.0303
        )  # Simple-to-complex transition scaling exponent for silicate rock
        CXEXPI = 0.155  # Complex crater scaling exponent for ice
        SIMCOMKI = 3081.39  # Simple-to-complex transition scaling coefficient for ice
        SIMCOMPI = -1.22486  # Simple-to-complex transition scaling exponent for ice

        # These terms are used to compute the ratio of the transient crater to simple crater size
        simple_enlargement = 0.84  # See Melosh (1989) pg. 129 just following eq. 8.2.1

        # These terms are used in the exponent in the final rim radius/ simple crater radius vs  final radius / transition radius relationship
        # See Holsapple (1993) eq. 28
        complex_enlargement_factor = 1.02

        # These terms are used to compute the transition diameter as a function of gravity
        # The transition values come from CTEM and are a synthesis of Pike (1980), Croft (1985), Schenk et al. (2004).
        if self.target.transition_scale_type == "silicate":
            simple_complex_exp = SIMCOMPS
            simple_complex_fac = SIMCOMKS
            final_exp = CXEXPS
        elif self.target.transition_scale_type == "ice":
            simple_complex_exp = SIMCOMPI
            simple_complex_fac = SIMCOMKI
            final_exp = CXEXPI

        # The nominal value will be used for determining the range of the "transitional" morphology type
        self._transition_nominal = float(
            simple_complex_fac * self.target.gravity**simple_complex_exp
        )

        simple_enlargement_factor = 1.0 / simple_enlargement
        self._simple_enlargement_factor = float(simple_enlargement_factor)
        self._complex_enlargement_factor = float(complex_enlargement_factor)
        self._final_exp = float(final_exp)
        self._transition_diameter = float(self._transition_nominal)
        return



from __future__ import annotations

from typing import Any

import numpy as np
from astropy.constants import G

from cratermaker.constants import FloatLike
from cratermaker.utils.component_utils import ComponentBase, import_components
from cratermaker.utils.general_utils import (
    _create_catalogue,
    _set_properties,
    format_large_units,
    parameter,
)


class Target(ComponentBase):
    """
    Represents the target body in a crater simulation.

    This class encapsulates the properties of the target that is impacted, including
    its material composition, size, and other relevant physical characteristics.
    """

    _catalogue_header = [
        "name",
        "radius",
        "mass",
        "material",
        "transition_scale_type",
    ]
    # The catalogue was created with Swiftest
    _catalogue_values = [
        ("Mercury", 2439.40e3, 3.301001e23, "Soft Rock", "silicate"),
        ("Venus", 6051.84e3, 4.867306e24, "Hard Rock", "silicate"),
        ("Earth", 6371.01e3, 5.972168e24, "Wet Soil", "silicate"),
        ("Moon", 1737.53e3, 7.345789e22, "Soft Rock", "silicate"),
        ("Mars", 3389.92e3, 6.416909e23, "Soft Rock", "silicate"),
        ("Phobos", 11.17e3, 1.080000e16, "Soft Rock", "silicate"),
        ("Deimos", 6.30e3, 1.800000e15, "Soft Rock", "silicate"),
        ("Ceres", 469.70e3, 9.383516e20, "Soft Rock", "ice"),
        ("Vesta", 262.70e3, 2.590270e20, "Soft Rock", "silicate"),
        ("Io", 1821.49e3, 8.929649e22, "Hard Rock", "silicate"),
        ("Europa", 1560.80e3, 4.798574e22, "Ice", "ice"),
        ("Ganymede", 2631.20e3, 1.481479e23, "Ice", "ice"),
        ("Callisto", 2410.30e3, 1.075661e23, "Ice", "ice"),
        ("Titan", 2575.50e3, 1.345181e23, "Ice", "ice"),
        ("Rhea", 764.50e3, 2.306459e21, "Ice", "ice"),
        ("Dione", 562.50e3, 1.095486e21, "Ice", "ice"),
        ("Tethys", 536.30e3, 6.174430e20, "Ice", "ice"),
        ("Enceladus", 252.30e3, 1.080318e20, "Ice", "ice"),
        ("Mimas", 198.80e3, 3.750939e19, "Ice", "ice"),
        ("Ariel", 578.90e3, 1.250019e21, "Ice", "ice"),
        ("Umbriel", 584.70e3, 1.279535e21, "Ice", "ice"),
        ("Titania", 788.90e3, 3.338178e21, "Ice", "ice"),
        ("Oberon", 761.40e3, 3.076577e21, "Ice", "ice"),
        ("Miranda", 235.70e3, 6.442623e19, "Ice", "ice"),
        ("Triton", 1352.60e3, 2.140292e22, "Ice", "ice"),
        ("Charon", 606.00e3, 1.589680e21, "Ice", "ice"),
        ("Pluto", 1188.30e3, 1.302498e22, "Ice", "ice"),
        ("Arrokoth", 9.13e3, 7.485000e14, "Ice", "ice"),
    ]
    _catalogue = _create_catalogue(_catalogue_header, _catalogue_values)

    _density_catalogue = {
        "Water": 1000.0,
        "Sand": 1750.0,
        "Dry Soil": 1500.0,
        "Wet Soil": 2000.0,
        "Soft Rock": 2250.0,
        "Hard Rock": 2500.0,
        "Ice": 900.0,
    }

    def __init__(
        self,
        name: str,
        radius: FloatLike | None = None,
        diameter: FloatLike | None = None,
        mass: FloatLike | None = None,
        transition_scale_type: str | None = None,
        material: str | None = None,
        density: FloatLike | None = None,
        **kwargs: Any,
    ):
        """
        Initialize the target object, setting properties from the provided arguments.

        Parameters
        ----------
        name : str or None
            Name of the target body.
        radius : FloatLike or None
            Radius of the target body in km.
        diameter : FloatLike or None
            Diameter of the target body in km.
        mass : FloatLike or None
            Mass of the target body in kg.
        transition_scale_type : str or None
            Simple-to-complex transition scaling to use for the surface (either "silicate" or "ice").
        material : str or None
            Name of the material composition of the target body.
        density : FloatLike or None
            Volumetric density of the surface of the target body in kg/m^3.

        **kwargs : Any
            Additional keyword argumments that could be set by the user.

        Notes
        -----
        - The `radius` and `diameter` parameters are mutually exclusive. Only one of them should be provided.
        - Parameters set explicitly using keyword arguments will override those drawn from the catalogue.
        """
        object.__setattr__(self, "_updating", False)  # guard against recursive updates
        super().__init__(**kwargs)
        object.__setattr__(self, "_name", None)
        object.__setattr__(self, "_radius", None)
        object.__setattr__(self, "_mass", None)
        object.__setattr__(self, "_transition_scale_type", None)
        object.__setattr__(self, "_material", None)
        object.__setattr__(self, "_density", None)

        # ensure that only either diamter of radius is passed
        size_values_set = sum(x is not None for x in [diameter, radius])
        if size_values_set > 1:
            raise ValueError("Only one of diameter or radius may be set")
        if diameter is not None:
            radius = diameter / 2.0

        catalogue = kwargs.pop("catalogue", self.__class__._catalogue)

        # Set properties for the Target object based on the arguments passed to the function
        _set_properties(
            self,
            name=name,
            radius=radius,
            mass=mass,
            material=material,
            catalogue=catalogue,
            density=density,
            transition_scale_type=transition_scale_type,
            **kwargs,
        )
        arg_check = sum(
            x is None
            for x in [self.name, self.diameter, self.mass, self.transition_scale_type]
        )
        if arg_check > 0:
            raise ValueError("Invalid Target")

        if self._density is None:
            if self.material in self._density_catalogue:
                self._density = self._density_catalogue[self.material]

    def __str__(self) -> str:
        diameter = format_large_units(self.diameter, quantity="length")
        escape_velocity = format_large_units(self.escape_velocity, quantity="velocity")
        return (
            f"<Target: {self.name}>\n"
            f"Material: {self.material}\n"
            f"Diameter: {diameter}\n"
            f"Mass: {self.mass:.2e} kg\n"
            f"Surface density: {self.density:.1f} kg/m³\n"
            f"Transition Type: {self.transition_scale_type}\n"
            f"Escape Velocity: {escape_velocity}\n"
            f"Gravity: {self.gravity:.3f} m/s²"
        )

    @parameter
    def radius(self) -> float | None:
        return self._radius

    @radius.setter
    def radius(self, value: FloatLike):
        if value is not None and value <= 0:
            raise ValueError("Radius must be positive")
        setattr(self, "_radius", float(value))

    @property
    def diameter(self) -> float | None:
        if self._radius is not None:
            return 2 * self._radius

    @property
    def mass(self) -> float | None:
        return self._mass

    @mass.setter
    def mass(self, value: FloatLike):
        if value is not None and value <= 0:
            raise ValueError("Mass must be positive")
        setattr(self, "_mass", float(value))

    @property
    def name(self):
        """
        The name of the target body.

        Returns
        -------
        str
        """
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("name must be a string or None")
        self._name = value

    @parameter
    def material(self):
        """
        The name of the material composition of the target body.

        Returns
        -------
        str
        """
        return self._material

    @material.setter
    def material(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("material must be a string or None")
        self._material = value

    @parameter
    def density(self):
        """
        The volumetric density of the surface of the target body in kg/m^3.

        Returns
        -------
        float
        """
        return self._density

    @density.setter
    def density(self, value):
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("density must be a numeric value or None")
            if value <= 0:
                raise ValueError("density must be a positive number")
            self._density = float(value)
        else:
            self._density = None
        return

    @property
    def catalogue(self) -> str:
        from cratermaker.utils.general_utils import format_large_units

        if self.__class__._catalogue is None:
            return "This Projectile component does not have a catalogue."
        lines = []
        # Use the self.__class__._catalogue_header list to make the header
        header = "|".join([f"{h:<11}" for h in self.__class__._catalogue_header])
        lines.append(header)
        lines.append(len(header) * "-")
        for name, entry in self.__class__._catalogue.items():
            line = ""
            for k, v in entry.items():
                if k == "radius":
                    val = format_large_units(v, quantity="length")
                elif k == "mass":
                    val = f"{v:.2e} kg"
                else:
                    val = v
                line += f"|{val:<11}"
            lines.append(f"{name:<11}{line}")
        return "\n".join(lines)

    @property
    def catalogue_key(self):
        """
        The key used to identify the property used as the key in a catalogue.
        """
        return "name"

    @property
    def transition_scale_type(self):
        """
        The type of simple-to-complex transition scaling used for the surface, either 'silicate' or 'ice'.

        Returns
        -------
        str
        """
        return self._transition_scale_type

    @transition_scale_type.setter
    def transition_scale_type(self, value):
        valid_types = ["silicate", "ice"]
        if value not in valid_types and value is not None:
            raise ValueError(
                f"Invalid transition_scale_type: {value}. Must be one of {valid_types} or None"
            )
        self._transition_scale_type = value

    # The following are computed properties based on radius and mass
    @property
    def escape_velocity(self) -> float:
        """
        Calculate the escape velocity for the target body in SI units.

        Returns
        -------
        float
            Escape velocity in m/s.
        """
        return np.sqrt(2 * self.radius * self.gravity)

    @property
    def gravity(self) -> float | None:
        """
        Calculate the gravitational acceleration at the surface of the target body in SI units.

        Returns
        -------
        float
            Gravitational acceleration in m/s^2.
        """
        return G.value * self.mass / (self.radius**2)

    @classmethod
    def maker(
        cls: type[Target],
        target: Target | str = "Moon",
        radius: FloatLike | None = None,
        diameter: FloatLike | None = None,
        mass: FloatLike | None = None,
        transition_scale_type: str | None = None,
        material: str | None = None,
        density: FloatLike | None = None,
        **kwargs: Any,
    ) -> Target:
        """
        Initialize the target object, setting properties from the provided arguments.

        Parameters
        ----------
        target : str, Target, or None
            Name of the target body or a Target object.
        radius : FloatLike or None
            Radius of the target body in km.
        diameter : FloatLike or None
            Diameter of the target body in km.
        mass : FloatLike or None
            Mass of the target body in kg.
        transition_scale_type : str or None
            Simple-to-complex transition scaling to use for the surface (either "silicate" or "ice").
        material : str or None
            Name of the material composition of the target body.
        density : FloatLike or None
            Volumetric density of the surface of the target body in kg/m^3.
        **kwargs : Any
            Additional keyword argumments that could be set by the user.

        Notes
        -----
        - The `radius` and `diameter` parameters are mutually exclusive. Only one of them should be provided.
        - Parameters set explicitly using keyword arguments will override those drawn from the catalogue.
        """

        if target is None:
            try:
                target = cls(
                    name="Moon",
                    radius=radius,
                    diameter=diameter,
                    mass=mass,
                    transition_scale_type=transition_scale_type,
                    material=material,
                    density=density,
                    **kwargs,
                )
            except Exception as e:
                raise RuntimeError("Error initializing target.") from e
        elif isinstance(target, str):
            try:
                target = cls(
                    name=target,
                    radius=radius,
                    diameter=diameter,
                    mass=mass,
                    transition_scale_type=transition_scale_type,
                    material=material,
                    density=density,
                    **kwargs,
                )
            except KeyError:
                raise ValueError(
                    f"Target '{target}' not found in the catalogue. Please provide a valid target name."
                )
        elif not isinstance(target, Target):
            raise TypeError("target must be a string or a Target object")
        target._component_name = target.name

        return target


import_components(__name__, __path__, ignore_private=True)



import json
import math
from pathlib import Path
from typing import Any

import numpy as np
from numpy.typing import NDArray
from scipy import fft

from cratermaker.components.morphology import Morphology
from cratermaker.components.morphology.simplemoon import SimpleMoon


@Morphology.register("du2025")
class Du2025(SimpleMoon):
    """
    An operations class for computing the morphology of a crater and applying it to a surface mesh. This uses the morphology model of Du et al. 2025a,b

    Parameters
    ----------
    **kwargs : Any
        Additional keyword arguments to be passed to internal functions.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def _get_1D_power_spectral_density(
        self, feature, psd_coef, num_psd_component_effec=100
    ) -> NDArray:
        """
        This method construct a 1D power spectral density.
        Coeffcients are from [#]_.

        Parameters
        ----------
        feature : string
            For a 1D feature, choose from ejecta, rim, and floor
        psd_coef : dict (.json)
            Coeffcients used to constract a 1D power spectral density
        num_psd_component_effec : int
            Only reconstrut the sine waves with wavelengths smaller than 2pi/num_psd_component_effec to improve computational efficiency

        References
        ----------
        .. [#] Du, J., Minton, D. A., Blevins, A. M., Fassett, C. I., & Huang, Y. H. (2024). Spectral analysis of the morphology of fresh lunar craters I: Rim crest, floor, and rim flank outlines. Journal of Geophysical Research: Planets, 129(11), e2024JE008357. https://doi.org/10.1029/2024JE008357

        """
        # read the psd_coef outside this function. this is temporary until we find a better way to read the psd_coef
        psd_file = Path(__file__).parent / "psd_coef.json"
        with open(psd_file) as f:
            psd_coef = json.load(f)
        # ------------------------------------------------------------------------------------------------------------------
        num_psd_component = 5000
        # ------------------------------------------------------------------------------------------------------------------
        if self.crater.final_diameter < psd_coef["1D"][feature]["slope_12"]["D_tie"]:
            slope_12 = (
                psd_coef["1D"][feature]["slope_12"]["k1"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["slope_12"]["b1"]
            )
            slope_12_sigma = psd_coef["1D"][feature]["slope_12"]["sigma_1"]
        else:
            slope_12 = (
                psd_coef["1D"][feature]["slope_12"]["k2"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["slope_12"]["b2"]
            )
            slope_12_sigma = psd_coef["1D"][feature]["slope_12"]["sigma_2"]
        if self.crater.final_diameter < psd_coef["1D"][feature]["bp2_x"]["D_tie"]:
            bp2_x = (
                psd_coef["1D"][feature]["bp2_x"]["k1"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp2_x"]["b1"]
            )
            bp2_x_sigma = psd_coef["1D"][feature]["bp2_x"]["sigma_1"]
        else:
            bp2_x = (
                psd_coef["1D"][feature]["bp2_x"]["k2"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp2_x"]["b2"]
            )
            bp2_x_sigma = psd_coef["1D"][feature]["bp2_x"]["sigma_2"]
        if self.crater.final_diameter < psd_coef["1D"][feature]["bp2_y"]["D_tie"]:
            bp2_y = (
                psd_coef["1D"][feature]["bp2_y"]["k1"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp2_y"]["b1"]
            )
            bp2_y_sigma = psd_coef["1D"][feature]["bp2_y"]["sigma_1"]
        else:
            bp2_y = (
                psd_coef["1D"][feature]["bp2_y"]["k2"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp2_y"]["b2"]
            )
            bp2_y_sigma = psd_coef["1D"][feature]["bp2_y"]["sigma_2"]
        if self.crater.final_diameter < psd_coef["1D"][feature]["bp3_y"]["D_tie"]:
            bp3_y = (
                psd_coef["1D"][feature]["bp3_y"]["k1"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp3_y"]["b1"]
            )
            bp3_y_sigma = psd_coef["1D"][feature]["bp3_y"]["sigma_1"]
        else:
            bp3_y = (
                psd_coef["1D"][feature]["bp3_y"]["k2"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp3_y"]["b2"]
            )
            bp3_y_sigma = psd_coef["1D"][feature]["bp3_y"]["sigma_2"]
        if self.crater.final_diameter < psd_coef["1D"][feature]["bp4_y"]["D_tie"]:
            bp4_y = (
                psd_coef["1D"][feature]["bp4_y"]["k1"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp4_y"]["b1"]
            )
            bp4_y_sigma = psd_coef["1D"][feature]["bp4_y"]["sigma_1"]
        else:
            bp4_y = (
                psd_coef["1D"][feature]["bp4_y"]["k2"] * self.crater.final_diameter
                + psd_coef["1D"][feature]["bp4_y"]["b2"]
            )
            bp4_y_sigma = psd_coef["1D"][feature]["bp4_y"]["sigma_2"]
        slope_12 = self.rng.normal(slope_12, slope_12_sigma)
        bp2_x = self.rng.normal(bp2_x, bp2_x_sigma)
        bp2_y = self.rng.normal(bp2_y, bp2_y_sigma)
        bp3_y = self.rng.normal(bp3_y, bp3_y_sigma)
        bp4_y = self.rng.normal(bp4_y, bp4_y_sigma)
        psd_sigma = psd_coef["1D"][feature]["psd_sigma"]
        # ------------------------------------------------------------------------------------------------------------------
        bp4_x = math.log10(2 * math.pi)
        bp3_x = math.log10(10**bp4_x / 2)
        interval = 2 * math.pi / num_psd_component
        freq = fft.fftfreq(num_psd_component, interval)
        wavelength = 1 / freq[1 : num_psd_component // 2]
        psd = np.zeros((num_psd_component // 2 - 1, 2))
        psd[:, 0] = wavelength
        # ----------------------------------------------------------------------------------------------------------------------
        bp2_x_index = int(10**bp4_x / 10**bp2_x - 1)
        k_23 = (bp3_y - bp2_y) / (bp3_x - bp2_x)
        b_23 = bp3_y - k_23 * bp3_x
        k_12 = slope_12
        b_12 = bp2_y - k_12 * bp2_x
        # -----------------------------------------------------------------------------------------------------------------------
        psd[0, 1] = 10**bp4_x
        psd[1, 1] = 10**bp3_y
        psd[2 : bp2_x_index + 1, 1] = 10 ** (
            k_23 * np.log10(psd[2 : bp2_x_index + 1, 0]) + b_23
        )
        psd[bp2_x_index + 1 :, 1] = 10 ** (
            k_12 * np.log10(psd[bp2_x_index + 1 :, 0]) + b_12
        )
        # ----------------------------------------------------------------------------------------------------------------------
        psd = psd[:num_psd_component_effec]
        psd_log = np.log10(psd[:, 1])
        psd_log += self.rng.normal(0, psd_sigma, psd_log.shape)
        psd[:, 1] = 10**psd_log
        psd = np.flipud(psd)
        return psd

    def _get_2D_power_spectral_density(
        self, feature, psd_coef, ejecta_radius_norm, floor_radius_norm, max_effec_freq
    ) -> tuple[NDArray, NDArray, NDArray, NDArray, NDArray, NDArray]:
        """
        This method constructs a 2D power spectral density.
        Coeffcients are from [#]_.

        Parameters
        ----------
        feature : string
            For a 2D feature, choose from ejecta, wall, and floor
        psd_coef : dict (.json)
            Coeffcients used to constract a 2D power spectral density
        ejecta_radius_norm : float
            The radius of the continuous ejecta normalized by the crater radius
        floor_radius_norm : float
            The radius of the floor normalized by the crater radius
        max_effec_freq : int
            Only reconstrut the sine waves with frequencies smaller than a maxmium effective frequency to improve computational efficiency

        References
        ----------
        .. [#] Du, J., Minton, D. A., Blevins, A. M., Fassett, C. I., & Huang, Y. H. (2025). Spectral Analysis of the Morphology of Fresh Lunar Craters II: Two-Dimensional Surface Elevations of the Continuous Ejecta, Wall, and Floor. Journal of Geophysical Research: Planets

        """
        # read the psd_coef outside this function. this is temporary until we find a better way to read the psd_coef
        psd_file = Path(__file__).parent / "psd_coef.json"
        with open(psd_file) as f:
            psd_coef = json.load(f)  # read in from _init_?
        # ------------------------------------------------------------------------------------------------------------------
        if feature == "ejecta":
            max_effec_freq = 40
        if feature == "wall":
            max_effec_freq = 40
        if feature == "floor":
            max_effec_freq = 100
        # ------------------------------------------------------------------------------------------------------------------
        if feature == "ejecta":
            theta_min = 0
            theta_max = 2 * math.pi
            r_min = 1
            r_max = ejecta_radius_norm
            area = (r_max - r_min) * (theta_max - theta_min)
            theta_number = 2000
            r_number = 400
        if feature == "wall":
            theta_min = 0
            theta_max = 2 * math.pi
            r_min = floor_radius_norm
            r_max = 1
            area = (r_max - r_min) * (theta_max - theta_min)
            theta_number = 1000
            r_number = 200
        if feature == "floor":
            r_min = -floor_radius_norm / math.sqrt(2)
            r_max = floor_radius_norm / math.sqrt(2)
            theta_min = r_min
            theta_max = r_max
            area = (r_max - r_min) ** 2
            theta_number = 512
            r_number = 512
        # ------------------------------------------------------------------------------------------------------------------
        side_r = np.linspace(0, r_max - r_min, num=r_number, endpoint=False)
        side_theta = np.linspace(
            0, theta_max - theta_min, num=theta_number, endpoint=False
        )
        grid_theta, grid_r = np.meshgrid(side_theta, side_r)
        side_freq_theta = np.fft.fftfreq(theta_number, side_theta[1] - side_theta[0])
        side_freq_r = np.fft.fftfreq(r_number, side_r[1] - side_r[0])
        freq_theta, freq_r = np.meshgrid(
            np.fft.fftshift(side_freq_theta), np.fft.fftshift(side_freq_r)
        )
        freq_theta_quadrant = freq_theta[r_number // 2 :, theta_number // 2 :]
        freq_r_quadrant = freq_r[r_number // 2 :, theta_number // 2 :]
        # ------------------------------------------------------------------------------------------------------------------
        if self.crater.final_diameter < psd_coef["2D"][feature]["p_max"]["D_tie"]:
            p_max = (
                psd_coef["2D"][feature]["p_max"]["k1"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["p_max"]["b1"]
            )
            p_max_sigma = psd_coef["2D"][feature]["p_max"]["sigma_1"]
        else:
            p_max = (
                psd_coef["2D"][feature]["p_max"]["k2"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["p_max"]["b2"]
            )
            p_max_sigma = psd_coef["2D"][feature]["p_max"]["sigma_2"]
        if self.crater.final_diameter < psd_coef["2D"][feature]["p_diff"]["D_tie"]:
            p_diff = (
                psd_coef["2D"][feature]["p_diff"]["k1"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["p_diff"]["b1"]
            )
            p_diff_sigma = psd_coef["2D"][feature]["p_diff"]["sigma_1"]
        else:
            p_diff = (
                psd_coef["2D"][feature]["p_diff"]["k2"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["p_diff"]["b2"]
            )
            p_diff_sigma = psd_coef["2D"][feature]["p_diff"]["sigma_2"]
        if self.crater.final_diameter < psd_coef["2D"][feature]["nu_fall"]["D_tie"]:
            nu_fall = (
                psd_coef["2D"][feature]["nu_fall"]["k1"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["nu_fall"]["b1"]
            )
            nu_fall_sigma = psd_coef["2D"][feature]["nu_fall"]["sigma_1"]
        else:
            nu_fall = (
                psd_coef["2D"][feature]["nu_fall"]["k2"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["nu_fall"]["b2"]
            )
            nu_fall_sigma = psd_coef["2D"][feature]["nu_fall"]["sigma_2"]
        if self.crater.final_diameter < psd_coef["2D"][feature]["psd_sigma"]["D_tie"]:
            psd_sigma = (
                psd_coef["2D"][feature]["psd_sigma"]["k1"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["psd_sigma"]["b1"]
            )
            psd_sigma_sigma = psd_coef["2D"][feature]["psd_sigma"]["sigma_1"]
        else:
            psd_sigma = (
                psd_coef["2D"][feature]["psd_sigma"]["k2"] * self.crater.final_diameter
                + psd_coef["2D"][feature]["psd_sigma"]["b2"]
            )
            psd_sigma_sigma = psd_coef["2D"][feature]["psd_sigma"]["sigma_2"]
        p_max = self.rng.normal(p_max, p_max_sigma)
        p_diff = self.rng.normal(p_diff, p_diff_sigma)
        nu_fall = self.rng.normal(nu_fall, nu_fall_sigma)
        psd_sigma = self.rng.normal(psd_sigma, psd_sigma_sigma)
        # ------------------------------------------------------------------------------------------------------------------
        freq_r_matrix = np.sqrt(freq_theta_quadrant**2 + freq_r_quadrant**2)
        freq_theta_matrix = np.arctan2(freq_r_quadrant, freq_theta_quadrant)
        if feature == "ejecta":
            if self.crater.final_diameter < psd_coef["2D"][feature]["E_rad"]["D_tie"]:
                E_rad = (
                    psd_coef["2D"][feature]["E_rad"]["k1"] * self.crater.final_diameter
                    + psd_coef["2D"][feature]["E_rad"]["b1"]
                )
                E_rad_sigma = psd_coef["2D"][feature]["E_rad"]["sigma_1"]
            else:
                E_rad = (
                    psd_coef["2D"][feature]["E_rad"]["k2"] * self.crater.final_diameter
                    + psd_coef["2D"][feature]["E_rad"]["b2"]
                )
                E_rad_sigma = psd_coef["2D"][feature]["E_rad"]["sigma_2"]
            E_rad = self.rng.normal(E_rad, E_rad_sigma)
            psd_log_quadrant = (
                p_diff
                / np.sqrt(1 - (E_rad * np.sin(freq_theta_matrix)) ** 2)
                * (np.exp(-np.sqrt(freq_r_matrix / nu_fall)) - 1)
                + p_max
            )
        if feature == "wall":
            if self.crater.final_diameter < psd_coef["2D"][feature]["E_circ"]["D_tie"]:
                E_circ = (
                    psd_coef["2D"][feature]["E_circ"]["k1"] * self.crater.final_diameter
                    + psd_coef["2D"][feature]["E_circ"]["b1"]
                )
                E_circ_sigma = psd_coef["2D"][feature]["E_circ"]["sigma_1"]
            else:
                E_circ = (
                    psd_coef["2D"][feature]["E_circ"]["k2"] * self.crater.final_diameter
                    + psd_coef["2D"][feature]["E_circ"]["b2"]
                )
                E_circ_sigma = psd_coef["2D"][feature]["E_circ"]["sigma_2"]
            E_circ = self.rng.normal(E_circ, E_circ_sigma)
            psd_log_quadrant = (
                p_diff
                / np.sqrt(1 - (E_circ * np.cos(freq_theta_matrix)) ** 2)
                * (np.exp(-np.sqrt(freq_r_matrix / nu_fall)) - 1)
                + p_max
            )
        if feature == "floor":
            psd_log_quadrant = (
                p_diff * (np.exp(-np.sqrt(freq_r_matrix / nu_fall)) - 1) + p_max
            )
        psd_log_quadrant_1st = self.rng.normal(psd_log_quadrant, psd_sigma)
        psd_log_quadrant_2nd = self.rng.normal(psd_log_quadrant, psd_sigma)
        # ------------------------------------------------------------------------------------------------------------------
        psd_log = np.zeros((r_number, theta_number))
        psd_log[r_number // 2 :, theta_number // 2 :] = psd_log_quadrant_1st
        psd_log[r_number // 2 + 1 :, 1 : theta_number // 2] = np.fliplr(
            psd_log_quadrant_2nd[1:, 1:]
        )
        psd_log[1 : r_number // 2, theta_number // 2 + 1 :] = np.flipud(
            psd_log_quadrant_2nd[1:, 1:]
        )
        psd_log[1 : r_number // 2, 1 : theta_number // 2] = np.fliplr(
            np.flipud(psd_log_quadrant_1st[1:, 1:])
        )
        # ---------------------------------------------------------------------------------------------------------------
        psd_log[1 : r_number // 2, theta_number // 2] = np.flip(
            psd_log[r_number // 2 + 1 :, theta_number // 2]
        )
        psd_log[r_number // 2, 1 : theta_number // 2] = np.flip(
            psd_log[r_number // 2, theta_number // 2 + 1 :]
        )
        # ---------------------------------------------------------------------------------------------------------------
        psd_log[0, theta_number // 2 + 1 :] = psd_log[1, theta_number // 2 + 1 :]
        psd_log[r_number // 2 + 1 :, 0] = psd_log[r_number // 2 + 1 :, 1]
        # ---------------------------------------------------------------------------------------------------------------
        psd_log[0, 1 : theta_number // 2] = np.flip(psd_log[0, theta_number // 2 + 1 :])
        psd_log[1 : r_number // 2 :, 0] = np.flip(psd_log[r_number // 2 + 1 :, 0])
        # --------------------------------------------------------------------------------------------------------------
        psd_log[0, 0] = psd_log[1, 1]
        psd_log[0, theta_number // 2] = psd_log[1, theta_number // 2]
        psd_log[r_number // 2, 0] = psd_log[r_number // 2, 1]
        # ------------------------------------------------------------------------------------------------------------------
        psd = 10**psd_log
        phase = self.rng.uniform(-math.pi, math.pi, size=(r_number, theta_number))
        if feature == "ejecta" or feature == "wall":
            psd[:, theta_number // 2] = np.zeros((r_number))
        # ------------------------------------------------------------------------------------------------------------------
        freq_theta_delta = (freq_theta.max() - freq_theta.min()) / theta_number
        freq_r_delta = (freq_r.max() - freq_r.min()) / r_number
        freq_theta_num = int(max_effec_freq / freq_theta_delta)
        freq_r_num = int(max_effec_freq / freq_r_delta)
        freq_theta = freq_theta[
            r_number // 2 - freq_r_num : r_number // 2 + freq_r_num,
            theta_number // 2 - freq_theta_num : theta_number // 2 + freq_theta_num,
        ]
        freq_r = freq_r[
            r_number // 2 - freq_r_num : r_number // 2 + freq_r_num,
            theta_number // 2 - freq_theta_num : theta_number // 2 + freq_theta_num,
        ]
        psd = psd[
            r_number // 2 - freq_r_num : r_number // 2 + freq_r_num,
            theta_number // 2 - freq_theta_num : theta_number // 2 + freq_theta_num,
        ]
        phase = phase[
            r_number // 2 - freq_r_num : r_number // 2 + freq_r_num,
            theta_number // 2 - freq_theta_num : theta_number // 2 + freq_theta_num,
        ]
        # ------------------------------------------------------------------------------------------------------------------
        psd = np.sqrt(psd * area)
        return psd, phase, freq_theta, freq_r, grid_theta, grid_r



"""
Plot a distribution of random velocities given a mean
=====================================================

.. rubric:: By David Minton

This example demonstrates the generation of random velocities using the :func:`get_random_velocity` function from the :ref:`montecarlo_utils <api-utils-montecarlo>` module. The expected distribution of imapct velocities is a modified Maxwell-Boltzmann distribution. The modification is related to the relationship between the mean velocity and the escape velocity. If vmean > vescape, then the distribution will adjusted so that the escape velocity is the minimum velocity by computing an encounter velocity then summing the encounter and escape velocities in quadrature. If vmean < vescape, then the distirbution will be a truncated maxwellian.

Here we sample 10000 velocities with a mean of 0.5 and 2x the escape velocity of the Moon, both with and without the escape velocity argument. Then we will compare a histogram of generated velocities with .

"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import maxwell

from cratermaker import Target
from cratermaker.utils.montecarlo_utils import get_random_velocity

target = Target.maker("Moon")
vescape = target.escape_velocity

size = 10000

factors = np.array([0.75, 2.0])
vmean_values = factors * vescape
titles = [f"vmean = {float(factor):0.2f} $\\times$ vescape" for factor in factors]

fig, axs = plt.subplots(2, 2, figsize=(12, 6), sharex="col", sharey="row")
for col, vmean in enumerate(vmean_values):
    v_no_escape = get_random_velocity(vmean=vmean, size=size)
    v_with_escape = get_random_velocity(vmean=vmean, size=size, vescape=vescape)

    for row, (velocities, label) in enumerate(
        [(v_no_escape, "No Escape Velocity"), (v_with_escape, "With Escape Velocity")]
    ):
        bins = np.linspace(0, vmean * 3 / vescape, 50)
        histogram, bins = np.histogram(velocities / vescape, bins=bins, density=True)
        bin_centers = 0.5 * (bins[1:] + bins[:-1])
        theoretical_dist = (
            maxwell.pdf(bin_centers * vescape, scale=vmean / np.sqrt(8 / np.pi))
            * vescape
        )

        ax = axs[row, col]
        ax.bar(
            bin_centers,
            histogram,
            width=bins[1] - bins[0],
            alpha=0.5,
            label=label,
        )
        ax.plot(
            bin_centers,
            theoretical_dist,
            label="Maxwell-Boltzmann",
            color="red",
        )
        rms = np.sqrt(np.mean(velocities**2)) / vescape
        ax.axvline(rms, color="orange", linestyle="--", label="RMS Velocity")
        ax.annotate(
            f"RMS = {rms:.2f}",
            xy=(rms, ax.get_ylim()[1] * 0.65),
            xytext=(5, 0),
            textcoords="offset points",
            rotation=90,
            va="top",
            ha="left",
            fontsize=12,
        )
        if col == 0:
            ax.set_ylabel("Probability Density")
        if row == 1:
            ax.set_xlabel("$V / V_{{escape}}$")
        ax.legend()
        if row == 0:
            ax.set_title(titles[col])

plt.tight_layout()
plt.show()



"""
The simple-to-complex transition diameter
=========================================

.. rubric:: By David Minton

This example demonstrates how to use the :ref:`Scaling <ug-scaling>` class to calculate the simple-to-complex transition diameter as a function of surface gravity bodies using the Monte Carlo scaling model. Because the scaling model is non-deterministic, we sample each body 100 times to get a distribution of transition diameters. We plot the mean and standard deviation of the transition diameter is plotted for each body, showing results that resemble Fig. 7 of Schenk et al. (2021) [#]_.

References
----------

.. [#] Schenk, P., Castillo-Rogez, J., Otto, K.A., Marchi, S., O’Brien, D., Bland, M., Hughson, K., Schmidt, B., Scully, J., Buczkowski, D., Krohn, K., Hoogenboom, T., Kramer, G., Bray, V., Neesemann, A., Hiesinger, H., Platz, T., De Sanctis, M.C., Schroeder, S., Le Corre, L., McFadden, L., Sykes, M., Raymond, C., Russell, C.T., (2021) Compositional control on impact crater formation on mid-sized planetary bodies: Dawn at Ceres and Vesta, Cassini at Saturn. Icarus 359, 114343. https://doi.org/10.1016/j.icarus.2021.114343

"""

import matplotlib.pyplot as plt
import numpy as np

from cratermaker import Scaling

silicate_bodies = ["Vesta", "Moon", "Mercury", "Mars", "Earth", "Venus"]
ice_bodies = [
    "Mimas",
    "Miranda",
    "Tethys",
    "Dione",
    "Rhea",
    "Ceres",
    "Ganymede",
    "Europa",
]

fig, ax = plt.subplots(figsize=(5, 5))
ax.set_xscale("log")
ax.set_yscale("log")
ax.set_xlabel("Surface Gravity [cm/s²]")
ax.set_ylabel("Transition Diameter [km]")
ax.set_xlim((4, 2000))
ax.set_ylim((0.6, 200))
color_index = 0
label_color_map = {}
nsamples = 100
bodies = silicate_bodies + ice_bodies
for body in bodies:
    scaling = Scaling.maker(target=body)
    Dt_list = []
    g_list = []
    for _ in range(nsamples):
        Dt_list.append(scaling.transition_diameter * 1e-3)
        g_list.append(scaling.target.gravity * 100)
        scaling.recompute()

    Dt_arr = np.array(Dt_list)
    g_arr = np.array(g_list)
    g_mean = np.mean(g_arr)
    Dt_mean = np.mean(Dt_arr)
    Dt_std = np.std(Dt_arr)
    if body == "Europa":
        label_offset = 0.6 * Dt_mean
    elif body == "Ceres":
        label_offset = 0.57 * Dt_mean
    elif body == "Rhea":
        label_offset = 0.53 * Dt_mean
    elif body == "Mercury" or body == "Venus":
        label_offset = 0.75 * Dt_mean
    else:
        label_offset = 0.5 * Dt_mean

    if body in silicate_bodies:
        marker = "o"
        va = "bottom"
        y_text = Dt_mean + label_offset
    else:
        marker = "^"
        va = "top"
        y_text = Dt_mean - label_offset

    ax.plot(g_arr, Dt_arr, marker, markersize=2, alpha=0.5, label=f"{body} samples")
    ax.errorbar(
        g_mean,
        Dt_mean,
        yerr=Dt_std,
        fmt=marker,
        color="k",
        capsize=3,
        label=f"{body} mean",
    )
    ax.text(g_mean, y_text, body, fontsize=9, ha="center", va=va)

plt.tight_layout()
plt.show()



"""
Plot isochrons for the Moon and Mars for 1 Ma, 1 Ga, and 4 Ga using the Neukum Production Function
==================================================================================================

.. rubric:: By Austin Blevins and David Minton

In this example, we will be using the "neukum" prodcution model in Cratermaker to plot isochrons for three different age surfaces. We will also format the plots so with similar axes as Figure 2 of Neukum, Ivanov, and Hartmann (2001) [#]_, but with a handy logarithmic grid.

The Neukum production function (NPF) is only defined over a limited range, from crater diameters of 0.01 km to 300 km. Because Cratermaker is intended to primarily be used for forward modeling of landscape evolution where regolith mixing by small impactors plays a significant role, we require a production function that can extend to much smaller diameters. Therefore we extrapolate the lower end of the NPF with a simple power law with a slope equal to that of the lower end.

The upper range of the NPF must also be extrapolated, but for a different reason. If we truncated the NPF at 300 km, this would artificially limit the size of the largest craters that could be produced in a simulation and simulations of the ancient lunar or martian surface would contain dozens of 300 km craters. To avoid this, we extrapolate the upper end of the NPF, however we cannot simply extrapolate the upper end with its own slope, as this would result in far too many very large craters. This is similar to the issue that was discussed in Minton et al. (2015) [#]_. Therefore we steepen the slope of the upper end by shifting the exponent by -2. Futher work would be needed to determine the best fit to the upper end of the NPF.

For this example, we plot the valid range of the NPF in solid black, and the extrapolations in orange with a dash-dot line style.

References
----------

.. [#] Neukum, G., Ivanov, B.A., Hartmann, W.K., (2001) Cratering Records in the Inner Solar System in Relation to the Lunar Reference System. Space Science Reviews 96, 55-86. `doi: 10.1023/A:1011989004263 <https://doi.org/10.1023/A:1011989004263>`
.. [#] Minton, D.A., Richardson, J.E., Fassett, C.I., (2015) Re-examining the main asteroid belt as the primary source of ancient lunar craters. Icarus 247, 172-190. `doi: 10.1016/j.icarus.2014.10.018 <https://doi.org/10.1016/j.icarus.2014.10.018>`

"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker

from cratermaker import Production

fig, axs = plt.subplots(1, 2, figsize=(8, 7))
axes = dict(zip(["Moon", "Mars"], axs))
tvals = [0.01, 1.0, 4.0]
x_min = 1e-3
x_max = 1e4
y_min = 1e-12
y_max = 1e6
nD = 1000
Dvals = np.logspace(np.log10(x_min), np.log10(x_max), num=nD)
for version, ax in axes.items():
    production = Production.maker("neukum", version=version)
    ax.title.set_text(version)
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_ylabel("$\\mathregular{N_{>D} (km^{-2})}$")
    ax.set_xlabel("Diameter (km)")
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)

    # This section will create a proper log-log grid
    ax.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.yaxis.set_minor_locator(
        ticker.LogLocator(base=10.0, subs=np.arange(2, 10), numticks=100)
    )
    ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
    ax.xaxis.set_minor_locator(
        ticker.LogLocator(base=10.0, subs=np.arange(2, 10), numticks=100)
    )
    ax.grid(True, which="minor", ls="-", lw=0.5, zorder=5)
    ax.grid(True, which="major", ls="-", lw=1, zorder=10)

    # Split the valid and extrapolated range to plot them in different line styles and colors
    # Cratermaker uses meters for length and My for time, so we need to convert to km and Ga
    # for the plots to match the figure in Neukum et al. (2001)
    inrange = (Dvals > production.sfd_range[0] * 1e-3) & (
        Dvals < production.sfd_range[1] * 1e-3
    )
    lo = Dvals < production.sfd_range[0] * 1e-3
    hi = Dvals > production.sfd_range[1] * 1e-3
    for t in tvals:
        Nvals = production.function(diameter=Dvals * 1e3, age=t * 1e3)
        Nvals = Nvals * 1e6
        ax.plot(
            Dvals[inrange], Nvals[inrange], "-", color="black", linewidth=1.0, zorder=50
        )
        ax.plot(Dvals[lo], Nvals[lo], "-.", color="orange", linewidth=2.0, zorder=50)
        ax.plot(Dvals[hi], Nvals[hi], "-.", color="orange", linewidth=2.0, zorder=50)
        labeli = int(0.25 * nD)
        ax.text(
            Dvals[labeli],
            3 * Nvals[labeli],
            f"{t:.2f} ",
            ha="left",
            va="top",
            rotation=-72,
        )

plt.tight_layout()
plt.show()



"""
Create a crater and ejecta profile with the "simplemoon" morphology model
=========================================================================

.. rubric:: By David Minton

This example showcases how to create a crater and ejecta profile using the "simplemoon" morphology model from the CraterMaker package. The crater is created with a radius of 1 km, and the profiles are plotted in a 2D space normalized to the crater radius.

"""

import matplotlib.pyplot as plt
import numpy as np

from cratermaker import Crater, Morphology

crater = Crater.maker(final_radius=1.0e3)
morphology = Morphology.maker("simplemoon")

rc = crater.final_radius
rvals = np.linspace(0, 3.0 * rc, 1000)
hvals = morphology.crater_profile(crater, rvals)
ejvals = morphology.ejecta_profile(crater, rvals)

# Filter out the interior of the crater and add the ejecta to the crater profile to draw the combined profile
rej = rvals[ejvals > 0]
ejvals = hvals[ejvals > 0] + ejvals[ejvals > 0]

fig, ax = plt.subplots(figsize=(8, 2))

ax.set_xlabel("Distance ($r/r_c$)")
ax.set_ylabel("Height ($h/r_c$)")

ax.set_aspect("equal")
ax.plot(rvals / rc, hvals / rc)
ax.plot(rej / rc, ejvals / rc)

plt.show()



from __future__ import annotations

import sys
from typing import (
    TYPE_CHECKING,
    Any,
)

try:
    if sys.version_info >= (3, 11):
        from typing import Self, TypeAlias
    else:
        from typing import TypeAlias

        from typing_extensions import Self
except ImportError:
    if TYPE_CHECKING:
        raise
    else:
        Self: Any = None

import math
from typing import TYPE_CHECKING

import numpy as np
from numpy.random import Generator

from cratermaker.constants import FloatLike
from cratermaker.utils import montecarlo_utils as mc
from cratermaker.utils.component_utils import ComponentBase, import_components
from cratermaker.utils.general_utils import (
    format_large_units,
    parameter,
    validate_and_normalize_location,
)

if TYPE_CHECKING:
    from cratermaker.components.target import Target


class Projectile(ComponentBase):
    _registry: dict[str, Projectile] = {}
    _catalogue = None

    def __init__(
        self,
        sample: bool = True,
        mean_velocity: FloatLike | None = None,
        velocity: FloatLike | None = None,
        density: FloatLike | None = None,
        angle: FloatLike | None = None,
        direction: FloatLike | None = None,
        location: tuple[float, float] | None = None,
        target: Target | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs,
    ):
        """
        This is the abstract base class for all projectile models. It defines the interface for generating projectile velocities, angles, and densities for a given target body.

        Parameters
        ----------
        sample : bool
            Flag that determines whether to sample impact velocities, angles, and directions from distributions. If set to True, the `mean_velocity` argument is required. If set to False, the `velocity` argument is required.
        mean_velocity : float, optional
            The mean velocity of the projectile in m/s. Required if `sample` is True, ignored if `sample` is False.
        velocity : float | None
            The impact velocity in m/s. If `sample` is True, this value is ignored. If `sample` is False, this value is required.
        density : float, optional
            The density of the projectile in kg/m^3.
        angle : float, optional
            The impact angle in degrees. Default is 90.0 degrees (vertical impact) if `sample` is False. If `sample` is True, this value is ignored.
        direction : float | None
            The impact direction in degrees. Default is 0.0 degrees (due North) if `sample` is False. If `sample` is True, this value is ignored.`
        location : tuple[float, float] | None
            The location of the projectile on the target body in (lon, lat) coordinates. If None, the location will be sampled from a distribution.
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.
        """
        from cratermaker.components.target import Target

        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

        object.__setattr__(self, "_sample", sample)
        object.__setattr__(self, "_mean_velocity", mean_velocity)
        object.__setattr__(self, "_velocity", velocity)
        object.__setattr__(self, "_density", density)
        object.__setattr__(self, "_angle", angle)
        object.__setattr__(self, "_direction", direction)
        object.__setattr__(self, "_location", location)
        object.__setattr__(self, "_target", Target.maker(target, **kwargs))

        if self.sample:
            if self.mean_velocity is None:
                raise ValueError("mean_velocity must be provided when sample is True")
        else:
            if self.velocity is None:
                raise ValueError("velocity must be provided when sample is False")
            if self.angle is None:
                raise ValueError("angle must be provided when sample is False")
            if self.direction is None:
                raise ValueError("direction must be provided when sample is False")
            if self.location is None:
                raise ValueError("location must be provided when sample is False")

        if self.density is None:
            raise ValueError("density must be provided")
        return

    def __str__(self) -> str:
        base = super().__str__()
        if self.sample:
            params = f"\nMean Velocity: {format_large_units(self.mean_velocity, quantity='velocity')}"
        else:
            params = (
                f"\nVelocity: {format_large_units(self.velocity, quantity='velocity')}"
            )
            params += f"\nAngle: {self.angle:.1f} degrees"
            params += f"\nDirection: {self.direction:.1f} degrees"
        return (
            f"{base}\n"
            f"Sample from distributions: {self.sample}\n"
            f"{params}\n"
            f"Density: {self.density:.1f} kg/m³\n"
        )

    def _copy(self, deep: bool = True, memo: dict[int, Any] | None = None) -> Self:
        import copy
        import inspect

        copier = copy.deepcopy if deep else copy.copy
        memo = {} if memo is None else memo

        # Get all init parameters except 'self'
        cls = self.__class__
        sig = inspect.signature(cls.__init__)
        init_keys = sig.parameters.keys() - {"self"}

        # Extract values for those keys
        init_kwargs = {}
        for k in init_keys:
            if k == "target":
                init_kwargs[k] = self.target
            elif k == "rng":
                init_kwargs[k] = self.rng
            else:
                attr = getattr(self, k, None)
                init_kwargs[k] = copier(attr, memo) if deep else copier(attr)

        return cls(**init_kwargs)

    def __copy__(self) -> Self:
        return self._copy(deep=False)

    def __deepcopy__(self, memo: dict[int, Any] | None = None) -> Self:
        return self._copy(deep=True, memo=memo)

    @classmethod
    def maker(
        cls,
        projectile: Projectile | str | None = None,
        mean_velocity: FloatLike | None = None,
        density: FloatLike | None = None,
        sample: bool = True,
        angle: FloatLike | None = None,
        velocity: FloatLike | None = None,
        direction: FloatLike | None = None,
        location: tuple[float, float] | None = None,
        target: Target | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ) -> Projectile:
        """
        Initialize an projectile model based on the provided name or class.

        Parameters
        ----------
        projectile : Projectile or str
            The projectile model to initialize. Can be a class or a string representing the model name.
        mean_velocity : float
            The mean velocity of the projectile in m/s.
        density : float
            The density of the projectile in kg/m^3.
        sample : bool
            Flag that determines whether to sample impact velocities, angles, and directions from distributions. If set to False, impact velocities will be set to the mean velocity, impact angles will be set to 90 degrees (vertical impact), and directions will be 0.
        angle : float
            The impact angle in degrees.
        velocity : float | None
            The impact velocity in m/s. If None, the velocity will be sampled from a distribution.
        direction : float | None
            The impact direction in degrees. If None, the direction will be sampled from a distribution.
        location : tuple[float, float]
            The location of the projectile on the target body in (lon, lat) coordinates. If None, the location willb e sampled from a distribution
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Projectile
            The initialized projectile model.

        Raises
        ------
        KeyError
            If the specified projectile model is not found in the registry.
        TypeError
            If the specified projectile model is not a string or a subclass of Projectile.
        """
        from cratermaker.components.projectile.asteroids import AsteroidProjectiles
        from cratermaker.components.projectile.comets import CometProjectiles
        from cratermaker.components.target import Target

        target = Target.maker(target, **kwargs)
        if projectile is None:
            if target.name in AsteroidProjectiles._catalogue:
                projectile = "asteroids"
            elif target.name in CometProjectiles._catalogue:
                projectile = "comets"
            else:
                projectile = "generic"
                mean_velocity = 20.0e3 if mean_velocity is None else mean_velocity

        # if this is a brand new uninstantiated projectile, we need to flag it so that it its properties can be set propertly. Otherwise, it should just pass through as is.
        isfresh = isinstance(projectile, str)

        projectile = super().maker(
            component=projectile,
            mean_velocity=mean_velocity,
            density=density,
            sample=sample,
            angle=angle,
            velocity=velocity,
            direction=direction,
            location=location,
            target=target,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )
        if isfresh:
            return projectile.new_projectile(
                velocity=velocity,
                angle=angle,
                direction=direction,
                location=location,
                **kwargs,
            )
        else:
            return projectile

    def new_projectile(
        self,
        velocity: FloatLike | None = None,
        angle: FloatLike | None = None,
        direction: FloatLike | None = None,
        location: tuple[float, float] | None = None,
        **kwargs: Any,
    ) -> Self:
        """
        Returns a new projectile instance with updated sampled or default values,
        based on the original instance.

        Parameters
        ----------
        velocity : float | None
            The impact velocity in m/s. If None, the velocity will be sampled from a distribution if `sample` is True, otherwise it will be set to the value of `mean_velocity`.
        angle : float, optional
            The impact angle in degrees. Default is 90.0 degrees (vertical impact) if `sample` is False.
        direction : float | None
            The impact direction in degrees. Default is 0.0 degrees (due North) if `sample` is False.
        location : tuple[float, float] | None
            The location of the projectile on the target body in (lon, lat) coordinates. If None, the location will be sampled from a distribution.
        **kwargs : Any
            Additional keyword arguments to override attributes.

        Returns
        -------
        Projectile
            A new Projectile instance with updated properties.
        """
        import copy

        new_obj = copy.copy(self)

        if velocity is not None:
            new_obj.velocity = velocity
        elif new_obj.sample:
            new_obj._velocity = float(
                mc.get_random_velocity(
                    vmean=new_obj.mean_velocity,
                    vescape=new_obj.target.escape_velocity,
                    rng=new_obj.rng,
                )[0]
            )
        elif new_obj._velocity is None:
            new_obj._velocity = new_obj.mean_velocity

        if angle is not None:
            new_obj.angle = angle
        elif new_obj.sample:
            new_obj._angle = float(mc.get_random_impact_angle(rng=new_obj.rng)[0])
        elif new_obj._angle is None:
            new_obj._angle = 90.0

        if direction is not None:
            new_obj.direction = direction
        elif new_obj.sample:
            new_obj._direction = float(
                mc.get_random_impact_direction(rng=new_obj.rng)[0]
            )
        elif new_obj._direction is None:
            new_obj._direction = 0.0

        if location is not None:
            new_obj.location = location
        elif new_obj.sample:
            new_obj._location = mc.get_random_location(rng=new_obj.rng)[0]
        elif new_obj._location is None:
            new_obj._location = (0, 0)

        return new_obj

    @parameter
    def sample(self):
        """
        Flag that determines whether to sample velocities, angles, and directions from distributions. If set to False, impact velocities will be set to the mean velocity, impact angles will be set to 90 degrees (vertical impact), and directions will be 0.

        Returns
        -------
        bool
        """
        return self._sample

    @sample.setter
    def sample(self, value):
        if not isinstance(value, bool):
            raise TypeError("sample must be a boolean value")
        self._sample = value
        return

    @parameter
    def mean_velocity(self):
        """
        The mean velocity of the projectile in m/s.

        Returns
        -------
        float
        """
        return self._mean_velocity

    @mean_velocity.setter
    def mean_velocity(self, value):
        if isinstance(value, np.ndarray):
            value = value.item()
        if isinstance(value, FloatLike):
            if value < 0:
                raise ValueError("mean_velocity must be a positive number")
            self._mean_velocity = float(value)
        else:
            raise TypeError("mean_velocity must be a numeric value")
        return

    @property
    def angle(self):
        """
        The impact angle in degrees.

        Returns
        -------
        float
        """
        return self._angle

    @angle.setter
    def angle(self, value):
        if isinstance(value, np.ndarray):
            value = value.item()
        if not isinstance(value, FloatLike):
            raise TypeError("angle must be a numeric value")
        if value < 0:
            raise ValueError("angle must be a positive number")
        self._angle = float(value)

    @property
    def direction(self):
        """
        The impact direction in degrees.

        Returns
        -------
        float
        """
        return self._direction

    @direction.setter
    def direction(self, value):
        if isinstance(value, np.ndarray):
            value = value.item()
        if not isinstance(value, FloatLike):
            raise TypeError("direction must be a numeric value")
        if value < 0:
            raise ValueError("direction must be a positive number")
        self._direction = float(value)

    @property
    def location(self) -> tuple[float, float]:
        """
        The location of the projectile on the target body.

        Returns
        -------
        tuple
            A tuple containing the (lon,lat) coordinates of the projectile in degrees.
        """
        return self._location

    @location.setter
    def location(self, value: tuple[float, float]):
        self._location = validate_and_normalize_location(value)

    @property
    def velocity(self):
        """
        The impact velocity in m/s.

        Returns
        -------
        float
        """
        return self._velocity

    @velocity.setter
    def velocity(self, value):
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("velocity must be a numeric value")
            if value < 0:
                raise ValueError("velocity must be a positive number")
            self._velocity = float(value)
        else:
            self._velocity = None
        return

    @parameter
    def density(self):
        """
        The density of the projectile in kg/m^3.

        Returns
        -------
        float
        """
        return self._density

    @density.setter
    def density(self, value):
        """
        Sets the density of the projectile in kg/m^3.

        Parameters
        ----------
        value : float | None
            The density in kg/m^3 or None to use a default value.

        Raises
        ------
        ValueError
            If the provided value is negative or None.
        TypeError
            If the provided value is not a numeric value or None.
        """
        if not isinstance(value, FloatLike):
            raise TypeError("density must be a numeric value")
        if value < 0:
            raise ValueError("density must be a positive number")
        self._density = float(value)

    @property
    def vertical_velocity(self):
        """
        The vertical component of the impact velocity in m/s.

        Returns
        -------
        float
        """
        return self.velocity * math.sin(math.radians(self.angle))

    @property
    def population(self):
        """
        The name of the population of the projectile model.

        Returns
        -------
        int
        """
        return self._component_name

    @property
    def target(self):
        """
        The target object for the projectile model.

        Returns
        -------
        Target
        """
        return self._target

    @target.setter
    def target(self, value):
        from cratermaker.components.target import Target

        self._target = Target.maker(value)

    @property
    def catalogue(self) -> str:
        from cratermaker.utils.general_utils import format_large_units

        if self.__class__._catalogue is None:
            return "This Projectile component does not have a catalogue."
        lines = []
        header1 = "Target"
        lines.append(f"\n{header1:<11}| Mean Velocity")
        lines.append(26 * "-")
        for name, velocity in self.__class__._catalogue.items():
            formatted_velocity = format_large_units(velocity, quantity="velocity")
            lines.append(f"{name:<11}|     {formatted_velocity}")
        return "\n".join(lines)


import_components(__name__, __path__, ignore_private=True)



import unittest

import numpy as np

from cratermaker.components.production import Production

production_models = Production.available()


class TestProduction(unittest.TestCase):
    def test_production_N_to_time(self):
        for model_name in production_models:
            production = Production.maker(production=model_name)
            for age_orig in np.logspace(np.log10(4500), np.log10(0.1), num=10):
                D = np.logspace(-1, 3, num=1000)
                N = production.function(diameter=D, age=age_orig)
                age_new = production.age_from_D_N(
                    cumulative_number_density=N, diameter=D
                )
                np.testing.assert_array_almost_equal(age_orig, age_new, decimal=2)

    def test_production_N_age_to_D(self):
        for model_name in production_models:
            production = Production.maker(production=model_name)
            for age in np.logspace(np.log10(4500), np.log10(0.1), num=10):
                D_orig = np.logspace(-1, 3, num=1000)
                N = production.function(diameter=D_orig, age=age)
                D_new = production.D_from_N_age(cumulative_number_density=N, age=age)
                np.testing.assert_array_almost_equal(D_orig, D_new, decimal=2)

    def test_sample_arguments(self):
        for model_name in production_models:
            production = Production.maker(production=model_name)

            # Test valid arguments
            try:
                production.sample(age=500.0, diameter_range=(10.0, 100.0), area=1e6)
            except ValueError as e:
                self.fail(
                    f"sample() raised ValueError unexpectedly with valid arguments! {e}"
                )

            # Test providing both age and cumulative_number
            with self.assertRaises(ValueError):
                production.sample(
                    age=500, diameter_number=(100, 10), diameter_range=(10, 100)
                )

            # Test providing both age_end and diameter_number_end
            with self.assertRaises(ValueError):
                production.sample(
                    age=1500,
                    age_end=1000,
                    diameter_number_end=(100, 10),
                    diameter_range=(10, 100),
                )

            # Test missing diameter_range
            with self.assertRaises(ValueError):
                production.sample(age=500)

            # Test diameter_range with invalid length
            with self.assertRaises(ValueError):
                production.sample(age=500, diameter_range=(10,), area=1000)

            # Test diameter_range with non-positive minimum diameter
            with self.assertRaises(ValueError):
                production.sample(age=500, diameter_range=(-1, 100), area=1000)

            # Test diameter_range with max diameter less than min diameter
            with self.assertRaises(ValueError):
                production.sample(age=500, diameter_range=(100, 10), area=1000)

            # Test invalid diameter_number length
            with self.assertRaises(ValueError):
                production.sample(diameter_number=(100,), diameter_range=(10, 100))

            # Test negative value in diameter_number
            with self.assertRaises(ValueError):
                production.sample(diameter_number=(-100, 10), diameter_range=(10, 100))

            # Test negative area
            with self.assertRaises(ValueError):
                production.sample(age=500, diameter_range=(10, 100), area=-1000)

            # Test non-scalar age
            with self.assertRaises(ValueError):
                production.sample(age=[500, 1000], diameter_range=(10, 100))

            # This checks if a ValueError is raised for an invalid length of diameter_number_end.
            with self.assertRaises(ValueError):
                production.sample(diameter_number_end=(100,), diameter_range=(10, 100))

            # Testing when a negative value is provided in diameter_number_end.
            with self.assertRaises(ValueError):
                production.sample(
                    diameter_number_end=(-100, 10), diameter_range=(10, 100)
                )

    def test_small_sample(self):
        neukum = Production.maker(production="neukum", version="Moon")
        # Test that the number of craters is zero when the age and area are ridiculously tiny
        diameter, age = neukum.sample(
            age=1e-6, diameter_range=(300e3, 1000e3), area=1e-6
        )
        self.assertEqual(diameter.size, 0)
        self.assertEqual(age.size, 0)

    def testmake_production(self):
        # Test that Production.maker returns a valid production model
        for model_name in production_models:
            production = Production.maker(production=model_name)
            self.assertIn(model_name, production_models)
            self.assertIsInstance(production, Production)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import xarray as xr
from numpy.typing import ArrayLike
from tqdm import tqdm

from cratermaker.constants import (
    _CIRCLE_FILE_NAME,
    _COMBINED_DATA_FILE_NAME,
    _EXPORT_DIR,
    _GRID_FILE_NAME,
    _SURFACE_DIR,
    _VTK_FILE_EXTENSION,
    FloatLike,
)

if TYPE_CHECKING:
    from cratermaker.components.surface import Surface


def to_vtk(
    surface: Surface,
    interval_number: int = 0,
    time_variables: dict | None = None,
    save_geometry=True,
    **kwargs,
) -> None:
    """
    Export the surface mesh to a VTK file and stores it in the default export directory.
    """
    from vtk import (
        VTK_POLYGON,
        vtkPoints,
        vtkUnstructuredGrid,
        vtkWarpScalar,
        vtkXMLPolyDataWriter,
    )
    from vtkmodules.util.numpy_support import numpy_to_vtk
    from vtkmodules.vtkFiltersCore import vtkPolyDataNormals
    from vtkmodules.vtkFiltersGeometry import vtkGeometryFilter

    from cratermaker import Surface

    if not isinstance(surface, Surface):
        raise TypeError("The surface argument must be an instance of the Surface class.")
    # Create the output directory if it doesn't exist
    out_dir = surface.simdir / _EXPORT_DIR
    out_dir.mkdir(parents=True, exist_ok=True)

    data_dir = surface.simdir / _SURFACE_DIR
    data_file_list = list(data_dir.glob("*.nc"))
    if surface.grid_file in data_file_list:
        data_file_list.remove(surface.grid_file)

    # Convert uxarray grid arrays to regular numpy arrays for vtk processing
    n_node = surface.n_node
    n_face = surface.n_face
    node_x = surface.node_x
    node_y = surface.node_y
    node_z = surface.node_z
    n_nodes_per_face = surface.n_nodes_per_face
    face_node_connectivity = surface.face_node_connectivity

    vtk_data = vtkUnstructuredGrid()
    nodes = vtkPoints()
    for i in range(n_node):
        nodes.InsertNextPoint(node_x[i], node_y[i], node_z[i])
    vtk_data.SetPoints(nodes)
    vtk_data.Allocate(n_face)
    for i, n in enumerate(n_nodes_per_face):
        point_ids = face_node_connectivity[i][0:n]
        vtk_data.InsertNextCell(VTK_POLYGON, n, point_ids)

    warp = vtkWarpScalar()
    warp.SetInputArrayToProcess(0, 0, 0, vtkUnstructuredGrid.FIELD_ASSOCIATION_POINTS, "node_elevation")

    writer = vtkXMLPolyDataWriter()
    writer.SetDataModeToBinary()
    writer.SetCompressorTypeToZLib()

    if save_geometry:
        # Saves the surface mesh and its geometry as a separate file
        geometry_variables = [
            "node_x",
            "node_y",
            "node_z",
            "node_lon",
            "node_lat",
            "face_x",
            "face_y",
            "face_z",
            "face_lon",
            "face_lat",
            "face_area",
            "face_size",
        ]
        current_grid = vtkUnstructuredGrid()
        current_grid.DeepCopy(vtk_data)

        for v in geometry_variables:
            # extract the attribute v from the surface object
            array = numpy_to_vtk(getattr(surface, v), deep=True)
            array.SetName(v)
            n = getattr(surface, v).size
            if n == surface.n_face:
                current_grid.GetCellData().AddArray(array)
            elif n == surface.n_node:
                current_grid.GetPointData().AddArray(array)

        geom_filter = vtkGeometryFilter()
        geom_filter.SetInputData(current_grid)
        geom_filter.Update()
        poly_data = geom_filter.GetOutput()

        normals_filter = vtkPolyDataNormals()
        normals_filter.SetInputData(poly_data)
        normals_filter.ComputeCellNormalsOn()
        normals_filter.ConsistencyOn()  # Tries to make normals consistent across shared edges
        normals_filter.AutoOrientNormalsOn()  # Attempt to orient normals consistently outward/inward
        normals_filter.SplittingOff()
        normals_filter.Update()
        poly_data_with_normals = normals_filter.GetOutput()

        output_filename = out_dir / _GRID_FILE_NAME.replace(".nc", f".{_VTK_FILE_EXTENSION}")
        writer.SetFileName(output_filename)
        writer.SetInputData(poly_data_with_normals)
        writer.Write()

    ds = surface.uxds.load()
    current_grid = vtkUnstructuredGrid()
    current_grid.DeepCopy(vtk_data)

    for v in ds.variables:
        array = numpy_to_vtk(ds[v].values, deep=True)
        array.SetName(v)
        n = ds[v].size
        if "n_face" in ds[v].dims:
            current_grid.GetCellData().AddArray(array)
        elif "n_node" in ds[v].dims:
            current_grid.GetPointData().AddArray(array)
            if v == "node_elevation":
                current_grid.GetPointData().SetActiveScalars(v)
        elif n == 1:
            current_grid.GetFieldData().AddArray(array)

    if time_variables is None:
        time_variables = {"elapsed_time": float(interval_number)}
    else:
        if not isinstance(time_variables, dict):
            raise TypeError("time_variables must be a dictionary")

    for k, v in time_variables.items():
        array = numpy_to_vtk(np.array([v]), deep=True)
        array.SetName(k)
        current_grid.GetFieldData().AddArray(array)

    geom_filter = vtkGeometryFilter()
    geom_filter.SetInputData(current_grid)
    geom_filter.Update()
    poly_data = geom_filter.GetOutput()

    normals_filter = vtkPolyDataNormals()
    normals_filter.SetInputData(poly_data)
    normals_filter.ComputeCellNormalsOn()
    normals_filter.ConsistencyOn()  # Tries to make normals consistent across shared edges
    normals_filter.AutoOrientNormalsOn()  # Attempt to orient normals consistently outward/inward
    normals_filter.SplittingOff()
    normals_filter.Update()
    poly_data_with_normals = normals_filter.GetOutput()

    warp.SetInputData(poly_data_with_normals)
    warp.Update()
    warped_output = warp.GetOutput()
    output_filename = out_dir / _COMBINED_DATA_FILE_NAME.replace(".nc", f"{interval_number:06d}.{_VTK_FILE_EXTENSION}")
    writer.SetFileName(output_filename)
    writer.SetInputData(warped_output)
    writer.Write()

    return


def make_circle_file(
    surface,
    diameters: FloatLike | ArrayLike,
    longitudes: FloatLike | ArrayLike,
    latitudes: FloatLike | ArrayLike,
    output_filename: str | Path | None = None,
    *args,
    **kwargs,
) -> None:
    """
    Plot circles of diameter D centered at the given location.

    Parameters
    ----------
    diameters : FloatLike or ArrayLike
        Diameters of the circles in m.
    longitudes : FloatLike or ArrayLike of Floats
        Longitudes of the circle centers in degrees.
    latitudes : FloatLike or ArrayLike of Floats
        Latitudes of the circle centers in degrees.
    out_filename : str or Path, optional
        Name of the output file. If not provided, the default is "circle.vtp"
    """
    from vtk import (
        vtkCellArray,
        vtkPoints,
        vtkpoly_data,
        vtkPolyLine,
        vtkXMLpoly_dataWriter,
    )

    if output_filename is None:
        output_filename = surface.simdir / _EXPORT_DIR / _CIRCLE_FILE_NAME
    else:
        output_filename = surface.simdir / _EXPORT_DIR / output_filename

    diameters = np.atleast_1d(diameters)
    longitudes = np.atleast_1d(longitudes)
    latitudes = np.atleast_1d(latitudes)
    # Check for length consistency
    if len(diameters) != len(longitudes) or len(diameters) != len(latitudes):
        raise ValueError("The diameters, latitudes, and longitudes, arguments must have the same length")

    # Validate non-negative values
    if np.any(diameters < 0):
        raise ValueError("All values in 'diameters' must be non-negative")

    sphere_radius = surface.target.radius

    def create_circle(lon, lat, circle_radius, num_points=360):
        """
        Create a circle on the sphere's surface with a given radius and center.

        Parameters
        ----------
        lon : float
            Longitude of the circle's center in degrees.
        lat : float
            Latitude of the circle's center in degrees.
        circle_radius : float
            Radius of the circle in meters.
        num_points : int, optional
            Number of points to use to approximate the circle. The default is 360.
        """
        # Create an array of angle steps for the circle
        radians = np.linspace(0, 2 * np.pi, num_points)
        # Convert latitude and longitude to radians
        lat_rad = np.deg2rad(lat)
        lon_rad = np.deg2rad(lon)

        # Calculate the Cartesian coordinates for the circle's center
        center_x = sphere_radius * np.cos(lat_rad) * np.cos(lon_rad)
        center_y = sphere_radius * np.cos(lat_rad) * np.sin(lon_rad)
        center_z = sphere_radius * np.sin(lat_rad)

        # Calculate the vectors for the local east and north directions on the sphere's surface
        east = np.array([-np.sin(lon_rad), np.cos(lon_rad), 0])
        north = np.array(
            [
                -np.cos(lon_rad) * np.sin(lat_rad),
                -np.sin(lon_rad) * np.sin(lat_rad),
                np.cos(lat_rad),
            ]
        )

        # Initialize arrays to hold the circle points
        x = np.zeros_like(radians)
        y = np.zeros_like(radians)
        z = np.zeros_like(radians)

        # Calculate the points around the circle
        for i in range(num_points):
            x[i] = center_x + circle_radius * np.cos(radians[i]) * east[0] + circle_radius * np.sin(radians[i]) * north[0]
            y[i] = center_y + circle_radius * np.cos(radians[i]) * east[1] + circle_radius * np.sin(radians[i]) * north[1]
            z[i] = center_z + circle_radius * np.cos(radians[i]) * east[2] + circle_radius * np.sin(radians[i]) * north[2]

        return x, y, z

    points = vtkPoints()
    lines = vtkCellArray()
    point_id = 0  # Keep track of the point ID across all circles

    for lon, lat, diameter in zip(longitudes, latitudes, diameters, strict=False):
        circle_radius = diameter / 2
        x, y, z = create_circle(lon, lat, circle_radius)

        for i in range(len(x)):
            points.InsertNextPoint(x[i], y[i], z[i])

        polyline = vtkPolyLine()
        polyline.GetPointIds().SetNumberOfIds(len(x))
        for i in range(len(x)):
            polyline.GetPointIds().SetId(i, point_id)
            point_id += 1

        lines.InsertNextCell(polyline)

    # Create a poly_data object and add points and lines to it
    poly_data = vtkpoly_data()
    poly_data.SetPoints(points)
    poly_data.SetLines(lines)

    # Write the poly_data to a VTK file
    writer = vtkXMLpoly_dataWriter()
    writer.SetFileName(output_filename)
    writer.SetInputData(poly_data)

    # Optional: set the data mode to binary to save disk space
    writer.SetDataModeToBinary()
    writer.Write()

    return
